/** @internal
 **
 **
 ** @file fbcollector.c
 ** IPFIX Collecting Process single transport session implementation
 **
 ** ------------------------------------------------------------------------
 ** Copyright (C) 2006-2008 Carnegie Mellon University. All Rights Reserved.
 ** ------------------------------------------------------------------------
 ** Authors: Brian Trammell <bht@cert.org>
 ** ------------------------------------------------------------------------
 ** GNU Lesser GPL Rights pursuant to Version 2.1, February 1999 
 ** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
 ** ------------------------------------------------------------------------
 */

#define _FIXBUF_SOURCE_
#include <fixbuf/private.h>

#include "fbcollector.h"

#ident "$Id: fbcollector.c 12302 2008-08-21 20:26:28Z inacio_svn $"

/*#################################################
 *
 * IPFIX functions for reading input, these are
 * the default functions
 *
 *#################################################*/


/**
 * fbCollectorDecodeMsgVL
 *
 * decodes the header of a variable length message to determine
 * how long the message is in order to read the appropriate
 * amount to complete the message
 *
 *
 * @return FALSE on error, TRUE on success
 */
static gboolean fbCollectorDecodeMsgVL(
    fbCollector_t               *collector,
    fbCollectorMsgVL_t          *hdr,
    size_t                      b_len,
    uint16_t                    *m_len,
    GError                      **err)
{
    uint16_t                    h_version;
    uint16_t                    h_len;

    /* collector is unused in this function*/
    (void)collector;
    
    h_version = g_ntohs(hdr->n_version);
    if (h_version != 0x000A) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IPFIX,
                    "Illegal IPFIX Message version 0x%04x; "
                    "input is probably not an IPFIX Message stream.",
                    g_ntohs(hdr->n_version));
        *m_len = 0;
        return FALSE;
    }
    
    h_len = g_ntohs(hdr->n_len);
    if (h_len < 16) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IPFIX,
                    "Illegal IPFIX Message length 0x%04x; "
                    "input is probably not an IPFIX Message stream.",
                    g_ntohs(hdr->n_len));
        *m_len = 0;
        return FALSE;
    }
    
    if (b_len && (h_len > b_len)) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_BUFSZ,
                    "Buffer too small to read IPFIX Message "
                    "(message size %hu, buffer size %lu).",
                    h_len, b_len);
        *m_len = 0;
        return FALSE;
    }
    
    *m_len = h_len;
    return TRUE;
}


/**
 * fbCollectorMessageHeaderNull
 *
 * this is used to process a PDU after it has been read in a message
 * based transport protocol (UDP, SCTP) to adjust the header if
 * needed before sending it to post read fixing.  This version does
 * nothing. NULL transform.
 *
 * @param collector pointer to the collector state structure
 * @param buffer pointer to the message buffer
 * @param b_len length of the buffer passed in
 * @param m_len pointer to the length of the resultant buffer
 * @param err pointer to a GLib error structure
 *
 * @return TRUE (this always works)
 *
 */
gboolean    fbCollectorMessageHeaderNull (
    fbCollector_t               *collector,
    uint8_t                     *buffer,
    size_t                      b_len,
    uint16_t                    *m_len,
    GError                      **err)
{
    *m_len = b_len;
    return TRUE; 
}



/**
 * fbCollectorPostProcNull
 *
 * this is used to process a PDU after it has been read in order to transform
 * it, except that this function does _no_ transforms
 *
 * @param collector _not used_
 *
 * @return TRUE (always succesfull)
 *
 */
static gboolean     fbCollectorPostProcNull(
    fbCollector_t   *collector,
    uint8_t         *dataBuf,
    size_t          *bufLen,
    GError          **err)
{
    (void)collector;
    (void)dataBuf;
    (void)bufLen;
    (void)err;
    
    return TRUE;
}


/**
 * fbCollectorCloseTranslatorNull
 *
 * default function to clean up the translator state, but there is
 * none, and this function does nothing
 *
 * @param collector current collector
 *
 */
static void         fbCollectorCloseTranslatorNull(
    fbCollector_t   *collector)
{
    (void)collector;
    
    return;
}




/*#################################################
 *
 * the rest of the meat of the collector implementation
 *
 *#################################################*/


/**
 * fbCollectorReadFile
 *
 *
 *
 */
static gboolean fbCollectorReadFile(
    fbCollector_t           *collector,
    uint8_t                 *msgbase,
    size_t                  *msglen,
    GError                  **err)
{
    int                     rc;
    uint16_t                h_len;
    gboolean                goodLen;
    
    /* Read and decode version and length */
    g_assert(*msglen > 4);
    rc = fread(msgbase, 1, 4, collector->stream.fp);
    if (rc > 0) {
        goodLen = collector->coreadLen(collector, (fbCollectorMsgVL_t *)msgbase,
                                        *msglen, &h_len, err);
        if (FALSE == goodLen) return FALSE;
        msgbase += 4;
    } else if (feof(collector->stream.fp)) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOF,
                    "End of file");
        return FALSE;
    } else {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                    "I/O error: %s", strerror(errno));
        return FALSE;
    }

    /* read rest of message */
    rc = fread(msgbase, 1, h_len - 4, collector->stream.fp);
    if (rc > 0) {
        *msglen = rc + 4;
        if (!collector->copostRead(collector, msgbase, msglen, err)) {
            return FALSE;
        }
        return TRUE;
    } else if (feof(collector->stream.fp)) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOF,
                    "End of file");
        return FALSE;
    } else {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                    "I/O error: %s", strerror(errno));
        return FALSE;
    }
}

/**
 * fbCollectorCloseFile
 *
 *
 *
 */
static void fbCollectorCloseFile(
    fbCollector_t   *collector)
{
    if (collector->stream.fp != stdin) {
        fclose(collector->stream.fp);
    }
    collector->active = FALSE;
}

/**
 * fbCollectorAllocFP
 *
 *
 *
 */
fbCollector_t *fbCollectorAllocFP(
    void            *ctx,
    FILE            *fp)
{
    fbCollector_t   *collector = NULL;
    
    /* Create a new collector */
    collector = g_slice_new0(fbCollector_t);
    
    /* Fill the collector in */
    collector->ctx = ctx;
    collector->stream.fp = fp;
    collector->bufferedStream = TRUE;
    collector->active = TRUE;
    collector->coread = fbCollectorReadFile;
    collector->copostRead = fbCollectorPostProcNull;
    collector->coreadLen = fbCollectorDecodeMsgVL;
    collector->comsgHeader = fbCollectorMessageHeaderNull;
    collector->cotransClose = fbCollectorCloseTranslatorNull;
    collector->translationActive = FALSE;
    
    /* All done */
    return collector;
}

/**
 * fbCollectorAllocFile
 *
 *
 *
 */
fbCollector_t *fbCollectorAllocFile(
    void            *ctx,
    const char      *path,
    GError          **err)
{
    fbCollector_t   *collector = NULL;
    FILE            *fp = NULL;

    /* check to see if we're opening stdin */
    if ((strlen(path) == 1) && (path[0] == '-')) 
    {
        /* don't open a terminal */
        if (isatty(fileno(stdin))) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                        "Refusing to open stdin terminal for collection");
            return NULL;
        }
        
        /* yep, stdin */
        fp = stdin;
    } else {
        /* nope, just a regular file; open it. */
        fp = fopen(path, "r");
    }
    
    /* check for error */
    if (!fp) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                    "Couldn't open %s for collection: %s",
                    path, strerror(errno));
        return NULL;
    }
    
    /* allocate a collector */
    collector = fbCollectorAllocFP(ctx, fp);
    
    /* set the file close function */
    collector->coclose = fbCollectorCloseFile;
    
    /* set the default collector function */
    collector->copostRead = fbCollectorPostProcNull;
    
    /* default translator cleanup function */
    collector->cotransClose = fbCollectorCloseTranslatorNull;
    
    /* set the default message read length function */
    collector->coreadLen = fbCollectorDecodeMsgVL;

    /* set the default message header transform function */
    collector->comsgHeader = fbCollectorMessageHeaderNull;

    /* mark the stream if it is a buffered file pointer */
    collector->bufferedStream = TRUE;
    
    /* set that a input translator is not in use */
    collector->cotransClose = FALSE;
    
    /* all done */
    return collector;
}

#if FB_ENABLE_SCTP

/**
 * fbCollectorReadSCTP
 *
 *
 *
 */
static gboolean fbCollectorReadSCTP(
    fbCollector_t   *collector,
    uint8_t         *msgbase,
    size_t          *msglen,
    GError          **err)
{
    uint16_t                msgSize;
    struct sockaddr         peer;
    socklen_t               peerlen = sizeof(peer);
    struct sctp_sndrcvinfo  sri;
    int                     sctp_flags = 0;
    int                     rc;
    
    rc = sctp_recvmsg(collector->stream.fd, msgbase, *msglen,
                      &peer, &peerlen, &sri, &sctp_flags);
    
    if (rc > 0) {
        if (!collector->comsgHeader(collector, msgbase, rc, &msgSize, err)) {
            return FALSE;
        }
        *msglen = msgSize;
        if (!collector->copostRead(collector, msgbase, msglen, err)) {
            return FALSE;
        }
        return TRUE;
    } else if (rc == 0) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOF,
                    "End of file");
        return FALSE;
    } else if (errno == EINTR) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NLREAD,
                    "SCTP read interrupt");
        return FALSE;
    } else {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                    "TCP I/O error: %s", strerror(errno));
        return FALSE;
    }
}
#endif /* FB_ENABLE_SCTP */

/**
 * fbCollectorReadTCP
 *
 *
 *
 */
static gboolean fbCollectorReadTCP(
    fbCollector_t   *collector,
    uint8_t         *msgbase,
    size_t          *msglen,
    GError          **err)
{
    int                     rc;
    uint16_t                h_len, rrem;
    gboolean                goodLen;
    
    /* Read and decode version and length */
    g_assert(*msglen > 4);
    rrem = 4;
    while (rrem) {
        rc = read(collector->stream.fd, msgbase, rrem);
        if (rc > 0) {
            rrem -= rc;
            msgbase += rc;
        } else if (rc == 0) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOF,
                        "End of file");
            return FALSE;
        } else if (errno == EINTR) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NLREAD,
                        "TCP read interrupt at message start");
            return FALSE;
        } else {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                        "TCP I/O error: %s", strerror(errno));
            return FALSE;
        }
    }
    goodLen = collector->coreadLen(collector, (fbCollectorMsgVL_t *)(msgbase - 4),
                                   *msglen, &h_len, err);
    if (FALSE == goodLen) return FALSE;
    
    /* read rest of message */
    rrem = h_len - 4;
    while (rrem) {
        rc = read(collector->stream.fd, msgbase, rrem);
        if (rc > 0) {
            rrem -= rc;
            msgbase += rc;
        } else if (rc == 0) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOF,
                        "End of file");
            return FALSE;
        } else if (errno == EINTR) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NLREAD,
                        "TCP read interrupt in message");
            return FALSE;
        } else {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                        "TCP I/O error: %s", strerror(errno));
            return FALSE;
        }
    }
    
    /* Post process, if needed and return message length from header. */
    *msglen = h_len;
    if (!collector->copostRead(collector, msgbase, msglen, err)) {
        return FALSE;
    }
    return TRUE;
}

/**
 * fbCollectorVerifyUDPPeer
 *
 *
 *
 */
static gboolean fbCollectorVerifyUDPPeer(
    fbCollector_t   *collector,
    struct sockaddr *from,
    socklen_t       fromlen,
    GError          **err)
{
    /* stash the address if we've not seen it before */
    /* compare the address if we have */
    /* simulate no data (NLREAD) if message is from wrong peer */
    /* for now always verify */
    return TRUE;
}

/**
 * fbCollectorReadUDP
 *
 *
 *
 */
static gboolean fbCollectorReadUDP(
    fbCollector_t   *collector,
    uint8_t         *msgbase,
    size_t          *msglen,
    GError          **err)
{
    uint16_t        msgSize;
    ssize_t         recvlen;
    union {
        struct sockaddr         so;
        struct sockaddr_in      ip4;
        struct sockaddr_in6     ip6;
    }                           peer;
    socklen_t                   peerlen;

    peerlen = sizeof(peer);
    recvlen = recvfrom(collector->stream.fd, msgbase, *msglen, 0,
                       (struct sockaddr *)&peer, &peerlen);

    if (!collector->comsgHeader(collector, msgbase, recvlen, &msgSize, err)) {
        return FALSE;
    }
    
    if (msgSize > 0) {
        *msglen = msgSize;
        /* 
         * FIXME this is a bit broken. What we're doing here is
         * ensuring we reject any message not from the _first_ peer
         * (thereby ensuring session integrity). To really do a UDP
         * Collecting Process _right_ (to the extent that such is
         * possible, heh...), we'd need to map peer addresses to
         * sessions, not sockets to CP sessions.
         *
         * Seems a lot of work for an insecure optional protocol.
         * So we do this, instead.
         */
        if (!fbCollectorVerifyUDPPeer(collector, &(peer.so), peerlen, err)) {
            return FALSE;
        }
        if (!collector->copostRead(collector, msgbase, msglen, err)) {
            return FALSE;
        }
        return TRUE;
    } else if (errno == EINTR || errno == EWOULDBLOCK) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NLREAD,
                        "UDP read interrupt or timeout");
        return FALSE;
    } else {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                    "UDP I/O error: %s", strerror(errno));
        return FALSE;
    }
}

/**
 * fbCollectorCloseSocket
 *
 *
 *
 */
static void fbCollectorCloseSocket(
    fbCollector_t   *collector)
{
    close(collector->stream.fd);
    collector->active = FALSE;
}

/**
 * fbCollectorAllocSocket
 *
 *
 *
 */
fbCollector_t *fbCollectorAllocSocket(
    fbListener_t    *listener,
    void            *ctx,
    int             fd,
    struct sockaddr *peer,
    size_t          peerlen)
{
    fbCollector_t   *collector = NULL;
    fbConnSpec_t    *spec = fbListenerGetConnSpec(listener);
    
    /* Create a new collector */
    collector = g_slice_new0(fbCollector_t);
    
    /* Fill it in */
    collector->listener = listener;
    collector->ctx = ctx;
    collector->stream.fd = fd;
    collector->bufferedStream = FALSE;
    collector->active = TRUE;
    collector->copostRead = fbCollectorPostProcNull;
    collector->coreadLen = fbCollectorDecodeMsgVL;
    collector->comsgHeader = fbCollectorMessageHeaderNull;
    collector->coclose = fbCollectorCloseSocket;
    collector->cotransClose = fbCollectorCloseTranslatorNull;
    collector->translationActive = FALSE;
    if (peerlen) {
        memcpy(&(collector->peer.so), peer, 
               (peerlen > sizeof(collector->peer)) ? 
                    sizeof(collector->peer) : peerlen);
    }

    /* Select a reader function */
    switch(spec->transport) {
#if FB_ENABLE_SCTP
    case FB_SCTP:
        collector->coread = fbCollectorReadSCTP;
        break;
#endif
    case FB_TCP:
        collector->coread = fbCollectorReadTCP;
        break;
    case FB_UDP:
        collector->coread = fbCollectorReadUDP;
        break;
    default:
        g_assert_not_reached();
    }
    
    /* All done */
    return collector;
}

#if HAVE_OPENSSL

/**
 * fbCollectorReadTLS
 *
 *
 *
 */
static gboolean fbCollectorReadTLS(
    fbCollector_t   *collector,
    uint8_t         *msgbase,
    size_t          *msglen,
    GError          **err)
{
    int                     rc;
    uint16_t                h_len, rrem;
    
    /* Read and decode version and length */
    g_assert(*msglen > 4);
    rrem = 4;
    while (rrem) {
        rc = SSL_read(collector->ssl, msgbase, rrem);
        if (rc > 0) {
            rrem -= rc;
            msgbase += rc;
        } else if (rc == 0) {
            /* FIXME this isn't _quite_ robust but it's good enough for now.
               we'll fix this when we do TLS/TCP stress testing. */
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOF,
                        "TLS connection shutdown");
            return FALSE;
        } else {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                        "TLS I/O error at message start: %s",
                        ERR_error_string(ERR_get_error(), NULL));
            while (ERR_get_error());
            return FALSE;
        }
    }
    h_len = collector->coreadLen(collector, 
                    (fbCollectorMsgVL_t *)(msgbase - 4), *msglen, err);
    if (!h_len) return FALSE;
    
    /* read rest of message */
    rrem = h_len - 4;
    while (rrem) {
        rc = SSL_read(collector->ssl, msgbase, rrem);
        if (rc > 0) {
            rrem -= rc;
            msgbase += rc;
        } else {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                        "TLS I/O error in message: %s",
                        ERR_error_string(ERR_get_error(), NULL));
            while (ERR_get_error());
            return FALSE;
        }
    }
    
    /* All done. Return message length from header. */
    *msglen = h_len;
    return TRUE;
}

/**
 * fbCollectorCloseTLS
 *
 *
 *
 */
static void fbCollectorCloseTLS(
    fbCollector_t   *collector)
{
    SSL_shutdown(collector->ssl);
    SSL_free(collector->ssl);
    collector->active = FALSE;
}

/**
 * fbCollectorOpenTLS
 *
 *
 *
 */
static gboolean fbCollectorOpenTLS(
    fbCollector_t   *collector,
    GError          **err)
{
    fbConnSpec_t    *spec = fbListenerGetConnSpec(collector->listener);
    BIO             *conn;
    gboolean        ok = TRUE;
    
    /* Initialize SSL context if necessary */
    if (!spec->vssl_ctx) {
        if (!fbConnSpecInitTLS(spec, TRUE, err)) {
            return FALSE;
        }
    }
   
    /* wrap a stream BIO around the opened socket */
    if (!(conn = BIO_new_socket(collector->stream.fd, 1))) {
        ok = FALSE;
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_CONN,
                    "couldn't wrap socket for TLS: %s", 
                    ERR_error_string(ERR_get_error(), NULL));
        while (ERR_get_error());
        goto end;
    }
    
    /* create SSL socket */
    if (!(collector->ssl = SSL_new((SSL_CTX *)spec->vssl_ctx))) {
        ok = FALSE;
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_CONN,
                    "couldnt create TLS socket: %s", 
                    ERR_error_string(ERR_get_error(), NULL));
        while (ERR_get_error());
        goto end;
    }
    
    /* accept SSL connection */
    SSL_set_accept_state(collector->ssl);
    SSL_set_bio(collector->ssl, conn, conn);
    SSL_set_mode(collector->ssl, SSL_MODE_AUTO_RETRY);   
    if (SSL_accept(collector->ssl) <= 0) {
        ok = FALSE;
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_CONN,
                    "couldn't accept on connected TLS socket: %s", 
                    ERR_error_string(ERR_get_error(), NULL));
        while (ERR_get_error());
        goto end;    
    }
    
    /* FIXME do post-connection verification */
    
end:
    if (!ok) {
        collector->active = FALSE;
        if (collector->ssl) {
            SSL_free(collector->ssl);
            collector->ssl = NULL;
        } else if (conn) {
            BIO_vfree(conn);
        }
    }
    return ok;
}

#if HAVE_OPENSSL_DTLS

/**
 * fbCollectorOpenDTLS
 *
 *
 *
 */
static gboolean fbCollectorOpenDTLS(
    fbCollector_t   *collector,
    GError          **err)
{
    fbConnSpec_t    *spec = fbListenerGetConnSpec(collector->listener);
    BIO             *conn;
    gboolean        ok = TRUE;
    
    /* Initialize SSL context if necessary */
    if (!spec->vssl_ctx) {
        if (!fbConnSpecInitTLS(spec, TRUE, err)) {
            return FALSE;
        }
    }
    
    /* wrap a stream BIO around the opened socket */
    if (!(conn = BIO_new_dgram(collector->stream.fd, 1))) {
        ok = FALSE;
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_CONN,
                    "couldn't wrap socket for TLS: %s", 
                    ERR_error_string(ERR_get_error(), NULL));
        while (ERR_get_error());
        goto end;
    }
    
    /* create SSL socket */
    if (!(collector->ssl = SSL_new((SSL_CTX *)spec->vssl_ctx))) {
        ok = FALSE;
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_CONN,
                    "couldnt create TLS socket: %s", 
                    ERR_error_string(ERR_get_error(), NULL));
        while (ERR_get_error());
        goto end;
    }
    
    /* Enable cookie exchange */
    SSL_set_options(collector->ssl, SSL_OP_COOKIE_EXCHANGE);

    /* accept SSL connection */
    SSL_set_bio(collector->ssl, conn, conn);
    SSL_set_accept_state(collector->ssl);
    SSL_set_mode(collector->ssl, SSL_MODE_AUTO_RETRY);   
    if (SSL_accept(collector->ssl) <= 0) {
        ok = FALSE;
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_CONN,
                    "couldn't accept on connected TLS socket: %s", 
                    ERR_error_string(ERR_get_error(), NULL));
        while (ERR_get_error());
        goto end;    
    }

    /* FIXME do post-connection verification */
    
end:
    if (!ok) {
        collector->active = FALSE;
        if (collector->ssl) {
            SSL_free(collector->ssl);
            collector->ssl = NULL;
        } else if (conn) {
            BIO_vfree(conn);
        }
    }
    return ok;
}

#endif /* HAVE_OPENSSL_DTLS */

/**
 * fbCollectorAllocTLS
 *
 *
 *
 */
fbCollector_t *fbCollectorAllocTLS(
    fbListener_t    *listener,
    void            *ctx,
    int             fd,
    struct sockaddr *peer,
    size_t          peerlen,
    GError          **err)
{
    gboolean        ok = TRUE;
    fbCollector_t   *collector = NULL;
    fbConnSpec_t    *spec = fbListenerGetConnSpec(listener);
    
    /* Create a new collector */
    collector = g_slice_new0(fbCollector_t);
    
    /* Fill it in */
    collector->listener = listener;
    collector->ctx = ctx;
    collector->stream.fd = fd;
    collector->bufferedStream = FALSE;
    collector->active = TRUE;
    collector->copostRead = fbCollectorPostProcNull;
    collector->coreadLen = fbCollectorDecodeMsgVL;
    collector->comsgHeader = fbCollectorMessageHeaderNull;
    collector->coread = fbCollectorReadTLS;
    collector->coclose = fbCollectorCloseTLS;
    collector->cotransClose = fbCollectorCloseTranslatorNull;
    collector->translationActive = FALSE;
    if (peerlen) {
        memcpy(&(collector->peer.so), peer, 
               (peerlen > sizeof(collector->peer)) ? 
                    sizeof(collector->peer) : peerlen);
    }


    /* Do TLS accept atop opened socket */
    switch (spec->transport) {
    case FB_TLS_TCP:
        ok = fbCollectorOpenTLS(collector, err);
        break;
#if HAVE_OPENSSL_DTLS
    case FB_DTLS_UDP:
#if HAVE_OPENSSL_DTLS_SCTP
    case FB_DTLS_SCTP:        
#endif
        ok = fbCollectorOpenDTLS(collector, err);
    break;
#endif
    default:
        g_assert_not_reached();
    }

    /* Nuke collector on TLS setup error */
    if (!ok) {
        g_slice_free(fbCollector_t, collector);
        return NULL;
    }
        
    /* All done */
    return collector;
}

#endif /* HAVE_OPENSSL */

/**
 * fbCollectMessage
 *
 *
 *
 */
gboolean        fbCollectMessage(
    fbCollector_t   *collector,
    uint8_t         *msgbase,
    size_t          *msglen,
    GError          **err)
{
    /* Ensure stream is open */
    if (!collector->active) return FALSE;

    /* Attempt to read message */
    if (collector->coread(collector, msgbase, msglen, err)) return TRUE;
    
    /* Read failure; signal error */
    return FALSE;
}

/**
 * fbCollectorGetContext
 *
 *
 *
 */
void          *fbCollectorGetContext(
    fbCollector_t   *collector)
{
    return collector->ctx;
}


/**
 * fbCollectorHasTranslator
 *
 * use this is check to see if a protocol translation
 * is in use for this collector.  Needed by the transcode
 * and IPFIX machinery to get rid of some error checks
 * which no longer apply.
 *
 * @param collector pointer to the collector state struct
 *
 * @return TRUE if a translator is in use, FALSE otherwise
 */
gboolean        fbCollectorHasTranslator(
    fbCollector_t   *collector)
{
    return collector->translationActive;
}

/**
 * fbCollectorGetFD
 *
 *
 *
 */
int             fbCollectorGetFD(
    fbCollector_t   *collector)
{
    return collector->stream.fd;
}

/**
 * fbCollectorClose
 *
 *
 *
 */
void            fbCollectorClose(
    fbCollector_t   *collector)
{
    if (collector->active && collector->coclose) collector->coclose(collector);

    if (collector->listener) {
        fbListenerRemove(collector->listener, collector->stream.fd);
    }
}

/**
 * fbCollectorFree
 *
 *
 *
 */
void            fbCollectorFree(
    fbCollector_t   *collector)
{
    collector->cotransClose(collector);
    fbCollectorClose(collector);
    g_slice_free(fbCollector_t, collector);
}


/**
 * fbCollectorClearTranslator
 *
 * @param collector the collector on which to remove
 *        the translator
 *
 * @return TRUE on success, FALSE on failure
 */
gboolean    fbCollectorClearTranslator(
    fbCollector_t   *collector,
    GError          **err)
{
    collector->cotransClose(collector);
    
    return TRUE;
}


/**
 * fbCollectorSetTranslator
 *
 * this sets the collector input to any
 * given translator
 *
 * @param collector the collector to apply the protocol
 *        convertor to
 * @param postProcFunc a function called after the read
 *        to do any post processing/conversion to turn
 *        the buffer into an IPFIX buffer
 * @param vlMessageFunc function to determine the
 *        amount needed to complete the next read
 * @param headerFunc function to transform the header after
 *        a block read before it is sent to the 
 *        postProcFunc (called when vlMessageFunc isn't)
 * @param trCloseFunc if anything is needed to be cleaned
 *        up in the translator when a collector is closed
 *        this function will be called before the collector
 *        is closed
 * @param opaque a void pointer to hold a translator
 *        specific state structure
 * @param err holds the glib based error message on
 *        error
 *
 * @return TRUE on success, FALSE on error
 */
gboolean    fbCollectorSetTranslator(
    fbCollector_t               *collector,
    fbCollectorPostProc_fn      postProcFunc,
    fbCollectorVLMessageSize_fn vlMessageFunc,
    fbCollectorMessageHeader_fn headerFunc,
    fbCollectorTransClose_fn    trCloseFunc,
    void                        *opaque,
    GError                      **err)
{
    if (NULL != collector->translatorState)
    {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_TRANSMISC,
            "Translator is already set on this collector, "
            "must be cleared first");
        return FALSE;
    }
    
    collector->copostRead = postProcFunc;
    collector->coreadLen = vlMessageFunc;
    collector->comsgHeader = headerFunc;
    collector->cotransClose = trCloseFunc;
    collector->translatorState = opaque;
    collector->translationActive = TRUE;

    return TRUE;
}


