/*@internal
 *
 * fblistener.c
 * IPFIX Collecting Process connection listener implementation
 *
 * ------------------------------------------------------------------------
 * Copyright (C) 2006-2008 Carnegie Mellon University. All Rights Reserved.
 * ------------------------------------------------------------------------
 * Authors: Brian Trammell <bht@cert.org>
 * ------------------------------------------------------------------------
 * GNU Lesser GPL Rights pursuant to Version 2.1, February 1999 
 * Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
 * ------------------------------------------------------------------------
 */

#define _FIXBUF_SOURCE_
#include <fixbuf/private.h>

#ident "$Id: fblistener.c 12302 2008-08-21 20:26:28Z inacio_svn $"


/**
 *
 * Understanding socket handling and collector construction within the
 * fixbuf listener:
 *
 * Error handling on connections in fixbuf is very different between
 * TCP and UDP connections.  The reasons behind this are partial
 * tied up within the IPFIX standard and its handling of UDP.
 * But within fixbuf, this needs to be understood in terms of how
 * collector's are created with respect to listener's.
 *
 * For both TCP and UDP, when a listener is created, a listening
 * socket is created with the listener.  (lsock)
 *
 * For UDP, the listener sock (lsock) is the socket used to create
 * the collector.  The collector gets created immediately, and the
 * collector is the structure that is associated with the fBuf
 * structure which actually handles PDU's.
 *
 * For TCP, the case is different.  The listening socket is used
 * primarily for the listenerWait call.  It is used as a socket
 * passed to select waiting for connection establishment.  Then
 * an accept call is made which creates a new socket handle.
 * That socket handle is used is to create the collector, and the
 * lsock handle is left only within the listener.
 *
 * When an error occurs, the normal usage of the API would be
 * to call fBufFree and call listenerWait again.  In the case
 * of TCP this works.  The library will wait for a new connection
 * to the listener lsock and create a new collector from a new
 * socket from the accept call.  For UDP, this will not work, and
 * the library will simply hang.  (Each lsock also has a 
 * corresponding set of pipes to detect interrupts) and the select
 * call will simply wait on the read pipe handle.
 *
 *
 */

struct fbListener_st {
    /** Connection specifier for passive socket. */
    fbConnSpec_t                *spec;
    /** Base session. Used for internal templates. */
    fbSession_t                 *session;
    /** Last buffer returned by fbListenerWait(). */
    fBuf_t                      *lastbuf;
    /** Passive socket file descriptor. */
    int                         lsock;
    /** 
     * Interrupt pipe read end file descriptor. 
     * Used to unblock a call to fbListenerWait(). 
     */
    int                         rip;
    /** 
     * Interrupt pipe write end file descriptor. 
     * Used to unblock a call to fbListenerWait(). 
     */
    int                         wip;
    /**
     * used to hold the handle to the collector for
     * this listener
     */
    fbCollector_t               *collectorHandle;
    /** 
     * File descriptor table. 
     * Maps file descriptors to active listener-managed buffer instances.
     */
    GHashTable                  *fdtab;
    /** 
     * Application initialization function. Allows the application
     * to bind internal context to a collector, and to reject connections
     * after accept() but before session setup. 
     */
    fbListenerAppInit_fn        appinit;
    /** Application free function. Frees storage allocated by appinit. */
    fbListenerAppFree_fn        appfree;
};

typedef struct fbListenerWaitFDSet_st {
    fd_set                      fds;
    int                         maxfd;
    fBuf_t                      *fbuf;
} fbListenerWaitFDSet_t;

/**
 * fbListenerTeardownSocket
 *
 *
 *
 *
 */
static void fbListenerTeardownSocket(
    fbListener_t                *listener)
{
    /* nuke interrupt pipe */
    if (listener->rip != -1) {
        close(listener->rip);
        listener->rip = -1;
    }
    
    if (listener->wip != -1) {
        close(listener->wip);
        listener->wip = -1;
    }
    
    /* nuke passive socket */
    if (listener->lsock != -1) {
        close(listener->lsock);
        listener->lsock = -1;
    }
}

/**
 *fbListenerInitSocket
 *
 *
 *
 *
 */
static gboolean fbListenerInitSocket(
    fbListener_t                *listener,
    GError                      **err)
{
    int                         pfd[2];
    struct addrinfo             *ai = NULL;
    
    /* Create interrupt pipe */
    if (pipe(pfd)) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                    "fbListener error creating interrupt pipe: %s",
                    strerror(errno));
        return FALSE;
    }
    listener->rip = pfd[0];
    listener->wip = pfd[1];

    /* Look up the passive socket address */
    if (!fbConnSpecLookupAI(listener->spec, TRUE, err)) {
        fbListenerTeardownSocket(listener);
        return FALSE;
    }
    
    ai = (struct addrinfo *)listener->spec->vai;
    
    /* Create the passive socket */
    do {
        /* 
         * Kludge for SCTP. addrinfo doesn't accept SCTP hints.
         */
#if FB_ENABLE_SCTP
        if ((listener->spec->transport == FB_SCTP) ||
            (listener->spec->transport == FB_DTLS_SCTP)) {
            ai->ai_socktype = SOCK_STREAM;
            ai->ai_protocol = IPPROTO_SCTP;
        }
#endif
        /* Create socket and bind it to the passive address */
        listener->lsock = socket(ai->ai_family, ai->ai_socktype, 
                                 ai->ai_protocol);
        if (listener->lsock < 0) continue;
        if (bind(listener->lsock, ai->ai_addr, ai->ai_addrlen) < 0) {
            close(listener->lsock); listener->lsock = -1; continue;
        }

        /* Listen only on socket and sequenced packet sockets */
        if ((ai->ai_socktype == SOCK_STREAM)
#ifdef SOCK_SEQPACKET
            || (ai->ai_socktype == SOCK_SEQPACKET)
#endif
            ) {
            if (listen(listener->lsock, 1) < 0) {
                close(listener->lsock); listener->lsock = -1; continue;
            }
        }
        break;
    } while ((ai = ai->ai_next));

    /* check for no listenable socket */
    if (ai == NULL) {
        fbListenerTeardownSocket(listener);
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_CONN,
                    "couldn't create socket listening to %s:%s: %s", 
                    listener->spec->host ? listener->spec->host : "*", 
                    listener->spec->svc, strerror(errno));
        return FALSE;
    }


    /* All done. */
    return TRUE;
}

/**
 *fbListenerInitUDPSocket
 *
 *
 *
 *
 */
static gboolean fbListenerInitUDPSocket(
    fbListener_t                *listener,
    GError                      **err)
{
    void                        *ctx = NULL;
    fbCollector_t               *collector = NULL;
    fBuf_t                      *fbuf = NULL;
    
    /* Simulate accept on UDP socket */
    
    /* Ask application for context */
    if (listener->appinit) {
        if (!listener->appinit(listener, &ctx, listener->lsock, NULL, 0, err)) {
            return FALSE;
        }
    }

    /* Create collector on UDP socket */
    switch (listener->spec->transport) {
    case FB_UDP:
        collector = fbCollectorAllocSocket(listener, ctx, 
                                           listener->lsock, NULL, 0);
        break;
#if HAVE_OPENSSL_DTLS
    case FB_DTLS_UDP:
        collector = fbCollectorAllocTLS(listener, ctx, 
                                        listener->lsock, NULL, 0, err);
        break;
#endif
    default:
        g_assert_not_reached();
    }
    
    /* Check for collector alloc error */
    if (!collector) return FALSE;
    
    /* Create a buffer with a cloned session around the collector */
    fbuf = fBufAllocForCollection(fbSessionClone(listener->session), collector);

    /* Add collector to the file descriptor table */
    g_hash_table_insert(listener->fdtab, 
                        GINT_TO_POINTER(listener->lsock), fbuf);
    
    /* No more passive socket */
    listener->lsock = -1;
    
    /* store the handle to the collector */
    listener->collectorHandle = collector;

    /* All done. */
    return TRUE;
}

/**
 *fbListenerAlloc
 *
 *
 *
 *
 */
fbListener_t *fbListenerAlloc(
    fbConnSpec_t                *spec,
    fbSession_t                 *session,
    fbListenerAppInit_fn        appinit,
    fbListenerAppFree_fn        appfree,
    GError                      **err)
{
    fbListener_t                *listener = NULL;

    /* Allocate a new listener */
    listener = g_slice_new0(fbListener_t);
    
    /* -1 for file descriptors means no fd */
    listener->lsock = -1;
    listener->rip = -1;
    listener->wip = -1;
    
    /* Fill in what we can */
    listener->spec = fbConnSpecCopy(spec);
    listener->session = session;
    listener->appinit = appinit;
    listener->appfree = appfree;
    
    /* allocate file descriptor table */
    listener->fdtab = g_hash_table_new(g_direct_hash, g_direct_equal);

    /* Do transport-specific initialization */
    switch (spec->transport) {
#if FB_ENABLE_SCTP
    case FB_SCTP:
#if HAVE_OPENSSL_DTLS_SCTP
    case FB_DTLS_SCTP:
#endif
#endif
    case FB_TCP:
#if HAVE_OPENSSL
    case FB_TLS_TCP:
#endif
        if (!fbListenerInitSocket(listener, err)) {
            goto err;
        }
        break;
    case FB_UDP:
#if HAVE_OPENSSL_DTLS
    case FB_DTLS_UDP:
#endif
        /* FIXME this may leak on socket setup error for UDP. */
        if (fbListenerInitSocket(listener, err)) {
            if (!fbListenerInitUDPSocket(listener, err)) {
                fbListenerTeardownSocket(listener);
                goto err;
            }
        } else {
            goto err;
        }
            
        break;
    default:
        g_assert_not_reached();
    }
    
    /* Return the initialized listener */
    return listener;
    
err:
    if (listener) {
        if (listener->fdtab) {
            g_hash_table_destroy(listener->fdtab);
        }
        
        g_slice_free(fbListener_t, listener);
    }
    
    /* No listener */
    return NULL;
}


/**
 * fbListenerFreeBuffer
 *
 *
 *
 *
 */
static void   fbListenerFreeBuffer(
    void                        *vfd,
    fBuf_t                      *fbuf,
    void                        *vignore)
{    
    /* free the buffer; this will close the socket. */
    fBufFree(fbuf);
}

/**
 *fbListenerFree
 *
 *
 *
 *
 */
void            fbListenerFree(
    fbListener_t                *listener)
{
    /* shut down passive socket */
    fbListenerTeardownSocket(listener);

    /* free any open buffers we may have */
    g_hash_table_foreach(listener->fdtab,
                        (GHFunc)fbListenerFreeBuffer, NULL);
                        
    /* free the listener table */
    g_hash_table_destroy(listener->fdtab);
    
    /* free the connection specifier */
    fbConnSpecFree(listener->spec);

    /* free the listener itself */
    g_slice_free(fbListener_t, listener);
}

/**
 *fbListenerWaitAddFD
 *
 *
 *
 *
 */
static void   fbListenerWaitAddFD(
    void                        *vfd,
    void                        *vignore,
    fbListenerWaitFDSet_t       *lfdset)
{
    int                         fd = GPOINTER_TO_INT(vfd);

    FD_SET(fd,&(lfdset->fds));
    if (fd > lfdset->maxfd) lfdset->maxfd = fd;
}

/**
 * fbListenerWaitSearch
 *
 *
 *
 *
 */
static void   fbListenerWaitSearch(
    void                        *vfd,
    void                        *fbuf,
    fbListenerWaitFDSet_t       *lfdset)
{
    int                         fd = GPOINTER_TO_INT(vfd);
    
    if (FD_ISSET(fd,&(lfdset->fds))) {
        lfdset->fbuf = fbuf;
    }
}


/**
 * fbListenerWaitAccept
 *
 *
 *
 *
 */
static fBuf_t *fbListenerWaitAccept(
    fbListener_t                *listener,
    GError                      **err)
{
    int                         asock;
    union {
        struct sockaddr         so;
        struct sockaddr_in      ip4;
        struct sockaddr_in6     ip6;
    }                           peer;
    socklen_t                   peerlen;
    void                        *ctx = NULL;
    fbCollector_t               *collector = NULL;
    fBuf_t                      *fbuf = NULL;
    
    /* Accept the connection */
    peerlen = sizeof(peer);
    asock = accept(listener->lsock, &(peer.so), &peerlen);
    if (asock < 0) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                    "listener accept error: %s", 
                    strerror(errno));
        return NULL;
    }

    /* Okay, we have a socket. Ask the application for context. */
    if (listener->appinit) {
        if (!listener->appinit(listener, &ctx, asock, 
                               &(peer.so), peerlen, err)) {
            close(asock);
            return NULL;
        }
    }
        
    /* Create a collector as appropriate */
    switch (listener->spec->transport) {
#if FB_ENABLE_SCTP
    case FB_SCTP:
#endif
    case FB_TCP:
        collector = fbCollectorAllocSocket(listener, ctx, asock, 
                                           &(peer.so), peerlen);
        break;
#if HAVE_OPENSSL
#if HAVE_OPENSSL_DTLS_SCTP
    case FB_DTLS_SCTP:
#endif
    case FB_TLS_TCP:
        collector = fbCollectorAllocTLS(listener, ctx, asock, 
                                        &(peer.so), peerlen, err);
        break;
#endif
    default:
        g_assert_not_reached();
    }
    
    /* Check for collector creation error */
    if (!collector) return NULL;
    
    /* Create a buffer with a cloned session around the collector */
    fbuf = fBufAllocForCollection(fbSessionClone(listener->session), collector);
    
    /* Make the buffer manual */
    fBufSetAutomaticMode(fbuf, FALSE);
    
    /* Add buffer to the file descriptor table */
    g_hash_table_insert(listener->fdtab, GINT_TO_POINTER(asock), fbuf);
    
    /* store the collector handle */
    listener->collectorHandle = collector;
    
    /* All done. */
    return fbuf;
}

/**
 * fbListenerRemove
 *
 *
 *
 *
 */
void fbListenerRemove(
    fbListener_t        *listener,
    int                 fd)
{
    /* remove buffer from fd */
    g_hash_table_remove(listener->fdtab, GINT_TO_POINTER(fd));
}

/**
 * fbListenerWait
 *
 *
 *
 *
 */
fBuf_t *fbListenerWait(
    fbListener_t                *listener,
    GError                      **err)
{
    fbListenerWaitFDSet_t       lfdset;
    uint8_t                     byte;
    int                         rc;

    /* set up the select call... */
    FD_ZERO(&lfdset.fds);
    lfdset.maxfd = 0;
    /* interrupt pipe read end */
    fbListenerWaitAddFD(GINT_TO_POINTER(listener->rip), NULL, &lfdset);
    /* listener socket if available */
    if (listener->lsock >= 0) {
        fbListenerWaitAddFD(GINT_TO_POINTER(listener->lsock), NULL, &lfdset);
    }
    /* any open collectors we may have */
    g_hash_table_foreach(listener->fdtab,
                        (GHFunc)fbListenerWaitAddFD, &lfdset);
    
    /* wait for data available on one of our file descriptors */
    rc = select(lfdset.maxfd + 1, &lfdset.fds, NULL, NULL, NULL);
    if (rc < 0) {
        if (errno == EINTR) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NLREAD,
                        "Interrupted listener wait");
            return NULL;            
        } else {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IO,
                        "listener wait error: %s", 
                        strerror(errno));
            return NULL;
        }
    }
    
    /* handle interrupt pipe read end */
    if (FD_ISSET(listener->rip, &lfdset.fds)) {
        /* consume and ignore return */
        read(listener->rip, &byte, sizeof(byte));
        /* throw error */
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NLREAD,
                    "External interrupt on pipe");
        return NULL;            
    }
    
    /* handle any pending accept, return the accepted buffer immediately. */
    if (listener->lsock >= 0 && 
        FD_ISSET(listener->lsock, &lfdset.fds)) {
        lfdset.fbuf = fbListenerWaitAccept(listener, err);
        if (!lfdset.fbuf) return NULL;
        listener->lastbuf = lfdset.fbuf;
        return lfdset.fbuf;
    }
    
    /* see if preferred collector buffer is ready */
    if (listener->lastbuf &&
        FD_ISSET(fbCollectorGetFD(fBufGetCollector(listener->lastbuf)), 
                                  &lfdset.fds)) {
        return listener->lastbuf;
    }
    
    /* return first available collector buffer */
    lfdset.fbuf = NULL;
    g_hash_table_foreach(listener->fdtab,
                         (GHFunc)fbListenerWaitSearch, &lfdset);
    listener->lastbuf = lfdset.fbuf;
    return lfdset.fbuf;
}

/**
 * fbListenerInterrupt
 *
 *
 *
 *
 */
void fbListenerInterrupt(
    fbListener_t        *listener)
{
    uint8_t             byte = 0xe7;
    
    /* write and ignore return */
    write(listener->wip, &byte, sizeof(byte));
}

/**
 * fbListenerGetConnSpec
 *
 *
 *
 *
 */
fbConnSpec_t        *fbListenerGetConnSpec(
    fbListener_t        *listener)
{
    return listener->spec;
}


/**
 *fbListenerGetCollector
 *
 * gets the collector allocated to the listener
 *
 */
gboolean            fbListenerGetCollector(
    fbListener_t        *listener,
    fbCollector_t       **collector,
    GError              **err)
{
    if (NULL == listener->collectorHandle) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_CONN,
            "no collector available to be retrieved");
        return FALSE;
    }
    
    *collector = listener->collectorHandle;
    
    return TRUE;
}
