/**
 *@internal
 *
 * @file fbnetflow.c
 *
 * This implements a Netflow convertor for translating into IPFIX
 * within the fixbuf structure
 *
 * ------------------------------------------------------------------------
 * Copyright (C) 2008 Carnegie Mellon University. All Rights Reserved.
 * ------------------------------------------------------------------------
 * Authors: Chris Inacio <inacio@cert.org>
 * ------------------------------------------------------------------------
 * GNU Lesser GPL Rights pursuant to Version 2.1, February 1999 
 * Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
 * ------------------------------------------------------------------------
 *
 * @Author:  $Author: inacio_svn $
 * @Date:    $Date: 2008-09-04 17:20:49 -0400 (Thu, 04 Sep 2008) $
 * @Version: $Revision: 12405 $
 *
 */

#define _FIXBUF_SOURCE_
#include <fixbuf/private.h>

#include "fbcollector.h"


#ident "$Id: fbnetflow.c 12405 2008-09-04 21:20:49Z inacio_svn $"

#ifndef FB_NETFLOW_DEBUG
#   define FB_NETFLOW_DEBUG 0
#endif

#if HAVE_ALIGNED_ACCESS_REQUIRED

/**
 * this set of macros for reading and writing U16's and U32's
 * uses memcpy's to avoid tripping over alignment issues on
 * platforms that cannot do unaligned access (e.g. SPARC, Alpha, 
 * etc).  The next section, after the else, does not use memcpy's
 * and operates just fine on architectures that don't crash
 * from an unaligned access (x86, PowerPC, etc.)
 */

/**
 * READU16INC
 *
 * read U16 and increment
 *
 * read a U16 value from ptr, it assumes the pointer
 * is properly aligned and pointing at the correct
 * place; and then increment ptr appropriately
 * according to its size to adjust for reading a
 * U16; increments it 2-bytes
 *
 * does network to host translation
 */
#define READU16INC(ptr,assignee) { \
        uint16_t *ru16_t16ptr = (uint16_t *)ptr; \
        uint16_t ru16_t16val = 0; \
        memcpy(&ru16_t16val, ru16_t16ptr, sizeof(uint16_t)); \
        assignee = g_ntohs(ru16_t16val); \
        ptr += sizeof(*ru16_t16ptr) / sizeof(*ptr); \
    }

#define READU16(ptr,assignee) { \
        uint16_t *ru16_t16ptr = (uint16_t *)ptr; \
        uint16_t ru16_t16val = 0; \
        memcpy(&ru16_t16val, ru16_t16ptr, sizeof(uint16_t)); \
        assignee = g_ntohs(ru16_t16val); \
    }

#define WRITEU16(ptr,value) { \
        uint16_t *ru16_t16ptr = (uint16_t *)ptr; \
        uint16_t ru16_t16val = g_htons(value); \
        memcpy(ru16_t16ptr, &ru16_t16val, sizeof(uint16_t)); \
    }


/**
 * READU32INC
 *
 * read U32 and increment ptr
 *
 * read a U32 from the ptr given in ptr, it assumes
 * it is aligned and positioned correctly, then
 * increment the ptr appropriately based on its
 * size and having read a U32 to increment it
 * 4-bytes ahead
 *
 * does network to host translation
 */
#define READU32INC(ptr,assignee) { \
    uint32_t *ru32_t32ptr = (uint32_t *)ptr; \
    uint32_t ru32_t32val = 0; \
    memcpy(&ru32_t32val, ru32_t32ptr, sizeof(uint32_t)); \
    assignee = g_ntohl(ru32_t32val); \
    ptr += sizeof(*ru32_t32ptr) / sizeof(*ptr); \
    }


#define READU32(ptr,assignee) { \
        uint32_t *ru32_t32ptr = (uint32_t *)ptr; \
        uint32_t ru32_t32val = 0; \
        memcpy(&ru32_t32val, ru32_t32ptr, sizeof(uint32_t)); \
        assignee = g_ntohl(ru32_t32val); \
    }

#define WRITEU32(ptr,value) { \
        uint32_t *ru32_t32ptr = (uint32_t *)ptr; \
        uint32_t ru32_t32val = g_htonl(value); \
        memcpy(ru32_t32ptr, &ru32_t32val, sizeof(uint32_t)); \
    }

#else

/**
 * READU16INC
 *
 * read U16 and increment
 *
 * read a U16 value from ptr, it assumes the pointer
 * is properly aligned and pointing at the correct
 * place; and then increment ptr appropriately
 * according to its size to adjust for reading a
 * U16; increments it 2-bytes
 *
 * does network to host translation
 */
#define READU16INC(ptr,assignee) { \
        uint16_t *ru16_t16ptr = (uint16_t *)ptr; \
        uint16_t ru16_t16val = 0; \
        ru16_t16val = g_ntohs(*ru16_t16ptr); \
        assignee = ru16_t16val; \
        ptr += sizeof(*ru16_t16ptr) / sizeof(*ptr); \
    }

#define READU16(ptr,assignee) { \
        uint16_t *ru16_t16ptr = (uint16_t *)ptr; \
        uint16_t ru16_t16val = 0; \
        ru16_t16val = g_ntohs(*ru16_t16ptr); \
        assignee = ru16_t16val; \
    }

#define WRITEU16(ptr,value) { \
        uint16_t *ru16_t16ptr = (uint16_t *)ptr; \
        *ru16_t16ptr = g_htons(value); \
    }


/**
 * READU32INC
 *
 * read U32 and increment ptr
 *
 * read a U32 from the ptr given in ptr, it assumes
 * it is aligned and positioned correctly, then
 * increment the ptr appropriately based on its
 * size and having read a U32 to increment it
 * 4-bytes ahead
 *
 * does network to host translation
 */
#define READU32INC(ptr,assignee) { \
    uint32_t *ru32_t32ptr = (uint32_t *)ptr; \
    uint32_t ru32_t32val = 0; \
    ru32_t32val = g_ntohl(*ru32_t32ptr); \
    assignee = ru32_t32val; \
    ptr += sizeof(*ru32_t32ptr) / sizeof(*ptr); \
    }


#define READU32(ptr,assignee) { \
        uint32_t *ru32_t32ptr = (uint32_t *)ptr; \
        uint32_t ru32_t32val = 0; \
        ru32_t32val = g_ntohl(*ru32_t32ptr); \
        assignee = ru32_t32val; \
    }

#define WRITEU32(ptr,value) { \
        uint32_t *ru32_t32ptr = (uint32_t *)ptr; \
        *ru32_t32ptr = g_htonl(value); \
    }
    
#endif


/** mini hash table for Netflow V9 */
typedef struct fbCollectorNetflowV9TemplateHash_st {
    /** id of the stored template, should be zeroed if not in use */
    uint16_t                    templateId;
    /** length of the template in question, zero is reserved for an unused field */
    uint16_t                    templateLength;    
    /** boolean flag set if template ID represents an options template */
    gboolean                    optionsTemplate;
} fbCollectorNetflowV9TemplateHash_t;


/** defines the extra state needed to convert from NetflowV9 to IPFIX */
struct fbCollectorNetflowV9State_st {
    uint32_t                    ipfixSeqNum;
    uint32_t                    netflowSeqNum;
    GHashTable                  *templateHash;
};

/*#################################################
 *
 * netflow v9 functions for the collector, used
 * to optionally read 
 *
 *#################################################*/

/**
 * fbCollectorDecodeV9MsgVL
 *
 * parses the header of a V9 message and determines
 * how much needs to be read in order to complete
 * the message, (at least in theory)
 *
 * @param collector a pointer to the collector state
 *        structure
 * @param hdr a pointer to the beginning of the buffer
 *        to parse as a message (get converted back 
 *        into a uint8_t* and used as such)
 * @param b_len length of the buffer passed in for the
 *        hdr
 * @param m_len length of the message header that still
 *        needs to be read (always set to zero, since 
 *        this reads the entire message)
 * @param err a pointer to glib error structure, used
 *        if an error occurs during processing the
 *        stream
 *
 *
 * @return number of octets to read to complete
 *         the message
 */
static gboolean     fbCollectorDecodeV9MsgVL(
     fbCollector_t               *collector,
     fbCollectorMsgVL_t          *hdr,
     size_t                      b_len,
     uint16_t                    *m_len,
     GError                      **err)
{
    uint16_t    recordCount;
    uint8_t     *dataBuf;
    uint8_t     *bufOffset;
    int         rc;
    struct setHeader_st {
        uint16_t    setId;
        uint16_t    setLength;  
    } *setHeaderPtr;
	unsigned int	loop;
	uint16_t		setLength;

    if (0x0009 != g_ntohs(hdr->n_version)) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
                    "Illegal NetflowV9 Message version 0x%04x; "
                    "input is probably not a NetflowV9 Message stream.",
                    g_ntohs(hdr->n_version));
        *m_len = 0;
        return FALSE;
    }

    /* so this is hopefully a netflow message;  I should now be able to
       read the entire message, it can't be larger than 1 packet, (the 
       question becomes, what's a packet size) */
    recordCount = htons(hdr->n_len);

    dataBuf = (uint8_t *)hdr;
    bufOffset = dataBuf + sizeof(hdr);


    /* read the rest of the v9 header, ugly trick ahead:
    gonna read in the uptime, and then throw away the read
    and finish reading the rest of the header, dumping the
    uptime, because we don't really want it */
    if (((bufOffset-dataBuf) + 16) < b_len) {
        g_set_error(err,FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
            "Error buffer to small to read NetflowV9 message header");
        *m_len = 0;
        return FALSE;
    }
    if (TRUE == collector->bufferedStream) {
        rc = fread(bufOffset, 1, 4, collector->stream.fp);
    } else {
        rc = read(collector->stream.fd, bufOffset, 4);
    }
    if (4 != rc) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
            "Could not complete read of the Netflow header(1)");
        *m_len = 0;
        return FALSE;
    }
    if (TRUE == collector->bufferedStream) {
        rc = fread(bufOffset, 1, 12, collector->stream.fp);
    } else {
        rc = read(collector->stream.fd, bufOffset, 12);
    }
    if (12 != rc) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
            "Could not complete read of the Netflow header(2)");
        *m_len = 0;
        return FALSE;
    }
    bufOffset += 12;

    /* so we don't really care about what is in the different sets,
       at this point, we just want to scan through recordCount
       number of them and read the length from each, and sum it,
       then we get to rewind the file offset so that we can go
       record-by-record back out to the application */
    for (loop = 0; loop < recordCount; loop++) {
        if (((bufOffset-dataBuf) + 4) < b_len) {
            g_set_error(err,FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
                "Error buffer to small to read NetflowV9 message");
            *m_len = 0;
            return FALSE;
        }
        if (TRUE == collector->bufferedStream) {
            rc = fread(bufOffset, 1, 4, collector->stream.fp);
        } else {
            rc = read(collector->stream.fd, bufOffset, 4);
        }
        if (4 != rc) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
                "Error reading set header in NetflowV9 message"
                "  expected read of 4 received %d", rc);
            *m_len = 0;
            return FALSE;
        }
        setHeaderPtr = (struct setHeader_st *)bufOffset;
        bufOffset += 4;
        setLength = g_ntohs(setHeaderPtr->setLength);
        if (((bufOffset-dataBuf) + setLength) < b_len) {
            g_set_error(err,FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
                "Error buffer to small to read NetflowV9 message");
            *m_len = 0;
            return FALSE;
        }
        if (TRUE == collector->bufferedStream) {
            rc = fread(bufOffset, 1, setLength, collector->stream.fp);
        } else {
            rc = read(collector->stream.fd, bufOffset, setLength);
        }
        if (setLength != rc) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
                "Error reading NetflowV9 set payload");
            *m_len = 0;
            return FALSE;
        }
        bufOffset += setLength;
    }

    
    *m_len = 0;
    return TRUE;
}

/**
 * fbCollectorMessageHeaderV9
 *
 * this converts a NetFlow V9 header into something every so slightly
 * closer to an IPFIX header; it dumps the up time with a big memcpy
 *
 * @param collector pointer to the collector state structure
 * @param buffer pointer to the message buffer
 * @param b_len length of the buffer passed in
 * @param m_len pointer to the length of the resultant buffer
 * @param err pointer to a GLib error structure
 *
 * @return TRUE on success, FALSE on error
 */
gboolean    fbCollectorMessageHeaderV9 (
    fbCollector_t               *collector,
    uint8_t                     *buffer,
    size_t                      b_len,
    uint16_t                    *m_len,
    GError                      **err)
{
    uint16_t tempRead16;
    
    /* first make sure the message seems like a NetFlow V9 message */
    tempRead16 = g_ntohs(*((uint16_t *)buffer));
    if (0x0009 != tempRead16) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
            "invalid version number for NetFlow V9, expecting 0x0009, received %#06x", tempRead16);
        return FALSE;
    }

    /* memcpy is no good here because src & dst overlap */
    memmove(buffer+4, buffer+8, b_len-8);
    
    /* return that we shortened the buffer */
    *m_len = b_len - 4;
    
    return TRUE; 
}



/**
 * netflowDataTemplateParse
 *
 * this parses a NetFlow V9 template and stores the results
 * into the template hash for this session.  It only stores
 * the template ID and the length of the resulting record.
 * it will error out on malformed templates and data records
 * which are not common between IPFIX and NetFlow V9
 *
 * @param collector pointer to the collector state record
 * @param dataBuf pointer to the buffer holding the template def
 *                points <b>after</b> the set ID and set length
 * @param recordLength pointer to the set header length field
 * @param msgBuf  pointer to the start of the netflow PDU
 * @param msgLen  pointer to the length of the whole msg
 * @param err GError pointer to store the error if one occurs
 *
 * @return TRUE on success, FALSE on error
 *
 */
static gboolean netflowDataTemplateParse (
    fbCollector_t   *collector, 
    uint8_t         *dataBuf, 
    uint16_t        *recordLength,
    uint8_t         *msgBuf,
    size_t          *msgLen,
    GError          **err)
{
    uint16_t templateId = 0;
    uint16_t fieldCount = 0;
    uint8_t  *bufPtr = dataBuf;
    uint16_t targetRecSize = 0;
    uint16_t recLength = g_ntohs(*recordLength);
    uint32_t bigTemplateId;
    unsigned int loop;
    struct fbCollectorNetflowV9State_st     *transState = 
        (struct fbCollectorNetflowV9State_st *)collector->translatorState;
    gpointer hashResult = NULL;
    struct fbCollectorNetflowV9TemplateHash_st *newTemplate = NULL;
    
    
    if ((recLength < 8) || 0 != (recLength%4)) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
            "NetFlow template record is either too short or not a multiple of "
            "4 octets, (recLength = %u)", recLength);
        return FALSE;
    }
    
    /* read the template ID */
    READU16INC(bufPtr, templateId);
    
    /* read the number of data records in the template */
    READU16INC(bufPtr, fieldCount);

    /* subtract 8 from the record length to account for the set
    header, (type & length 16-bits each) */
    if (fieldCount != ((recLength-8)/4)) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
            "NetFlow V9 field count does not equal remaing record length");
        return FALSE;
    }
    
    /* iterate through each type entry in the template
    record, make sure the IE model number is within 
    what we can handle and then record the length to
    build up the length of each template number */
    for (loop=0; loop < fieldCount; loop++) {
        uint16_t temp;
        uint8_t addReversePenFix = 0;
        
        READU16(bufPtr, temp);
        if (temp > 127 || 0 == temp) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
                "NetFlow V9 template data record model type is not "
                "common with IPFIX IE model");
            return FALSE;
        }

        /** because the IPFIX standard info model is broken, RFC 5102, we have
        to convert certain info model elements from their NetFlow v9 values
        into appropriate IPFIX numbers.  (Cisco will be none too pleased)
        */
        /* convert NetFlowV9 out bytes field into IPFIX octetTotalCount reverse*/
        if (23 == temp) {
            WRITEU16(bufPtr, (IPFIX_ENTERPRISE_BIT | 1));
            addReversePenFix = 1;
        }
        /* convert NetFlowV9 out pkts field into IPFIX packetTotalCount reverse */
        if (24 == temp) {
            WRITEU16(bufPtr, (IPFIX_ENTERPRISE_BIT | 2));
            addReversePenFix = 1;
        }
        bufPtr += sizeof(uint16_t);
        
        /* record how long each element is */
        READU16INC(bufPtr, temp);
        targetRecSize += temp;
        
        /* if we're supposed to add the reverse PEN for a NetFlow V9 -> IPFIX
        info model fix, do it now
        */
        if (0 != addReversePenFix) {
            if (FB_MSGLEN_MAX <= (*msgLen + sizeof(uint32_t))) {
                g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
                    "NetFlow V9 unable to convert information model "
                    "elements, no space");
                return FALSE;
            }
            /* make space for the extra 32-bit PEN */
            memmove(bufPtr+sizeof(uint32_t), bufPtr, 
                (*msgLen - (bufPtr - msgBuf)));
            /* write the reverse PEN into the template */
            WRITEU32(bufPtr, IPFIX_REVERSE_PEN);
            bufPtr += sizeof(uint32_t);
            /* update the length of this set in the set header */
            WRITEU16(recordLength, 
                (g_ntohs(*recordLength) + sizeof(uint32_t)));
            /* update the length that this record is; we error check this
               value later
            */
            recLength += sizeof(uint32_t);
            /* also increase the msglen, to note the change */
            *msgLen += sizeof(uint32_t);
        }
    }

    /* if there is extra stuff in the template, error out
    add 4 to compensate for the ptr offset of skipping the
    set header
    */
    if (((bufPtr-dataBuf)+4) != recLength) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
            "NetFlow V9 record length is longer than needed");
        return FALSE;
    }

    newTemplate = g_slice_new(fbCollectorNetflowV9TemplateHash_t);
    if (NULL == newTemplate) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
            "could not allocate template storage record");
        return FALSE;
    }
    newTemplate->templateId = templateId;
    newTemplate->templateLength = targetRecSize;
    newTemplate->optionsTemplate = FALSE;
    
    
    /* put the template into the hash, check/replace the template
    that is there if this template number already exists */
    bigTemplateId = (uint32_t)templateId;
    hashResult = g_hash_table_lookup(transState->templateHash, 
        (gconstpointer)bigTemplateId);
    if (NULL != hashResult) {
        g_hash_table_replace(transState->templateHash, 
            (gpointer)bigTemplateId, newTemplate);
    } else {
        g_hash_table_insert(transState->templateHash, 
            (gpointer)bigTemplateId, newTemplate);
    }

#if FB_NETFLOW_DEBUG == 1
    fprintf(stderr, "template inserted into hash: templateId %d,"
        " templateSize: %d\n", templateId, targetRecSize); /* debug */
#endif

    return TRUE;
}


/**
 * netflowOptionsTemplateParse
 *
 * this parses a NetFlow V9 options template and stores the results
 * into the template hash for this session.  It only stores
 * the template ID and the length of the resulting record.
 * it will error out on malformed templates
 *
 * @param collector pointer to the collector state record
 * @param dataBuf pointer to the buffer holding the template def
 *                points <b>after</b> the set ID and set length
 * @param recLength the length of the remainder of the template def
 * @param err GError pointer to store the error if one occurs
 *
 * @return TRUE on success, FALSE on error
 *
 */
static gboolean netflowOptionsTemplateParse (
    fbCollector_t   *collector, 
    uint8_t         *dataBuf, 
    uint16_t        recLength, 
    GError          **err)
{
    uint16_t templateId = 0;
    uint8_t *recOffset = dataBuf;
    uint16_t optScopeLen;
    uint16_t optLen;
    uint16_t templateLength = 0;
    uint16_t type;
    uint16_t fieldLen;
    uint32_t bigTemplateId;
    unsigned int loop;
    struct fbCollectorNetflowV9State_st     *transState = 
        (struct fbCollectorNetflowV9State_st *)collector->translatorState;
    gpointer hashResult = NULL;
    struct fbCollectorNetflowV9TemplateHash_st *newTemplate = NULL;
    
    
    /* read the template ID */
    READU16INC(recOffset,templateId);
    
    /* read the length of option scope bytes */
    READU16INC(recOffset,optScopeLen);
    
    /* read the length of the option bytes */
    READU16INC(recOffset,optLen);

    /* check the option scope length + option length to check to make sure
       that this record is sane */
    if ((recLength-10) - (optScopeLen + optLen) > 3) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
            "size mismatch in options template (recLength %u optScopeLen %u optLen %u)", recLength, optScopeLen, optLen);
        return FALSE;
    }

    for (loop = 0; loop < optScopeLen; loop += 4) {
        READU16INC(recOffset,type);
        READU16INC(recOffset,fieldLen);
        templateLength += fieldLen;
    }
    for (loop = 0; loop < optLen; loop += 4) {
        READU16INC(recOffset,type);
        READU16INC(recOffset,fieldLen);
        templateLength += fieldLen;
    }

    
    newTemplate = g_slice_new(fbCollectorNetflowV9TemplateHash_t);
    if (NULL == newTemplate) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
            "could not allocate template storage record");
        return FALSE;
    }
    newTemplate->templateId = templateId;
    newTemplate->templateLength = templateLength;
    newTemplate->optionsTemplate = TRUE;
    
    
    /* put the template into the hash, check/replace the template
    that is there if this template number already exists */
    bigTemplateId = (uint32_t)templateId;
    hashResult = g_hash_table_lookup(transState->templateHash, 
        (gconstpointer)bigTemplateId);
    if (NULL != hashResult) {
        g_hash_table_replace(transState->templateHash, 
            (gpointer)bigTemplateId, newTemplate);
    } else {
        g_hash_table_insert(transState->templateHash, 
            (gpointer)bigTemplateId, newTemplate);
    }
    
#if FB_NETFLOW_DEBUG == 1
    fprintf(stderr, "option template inserted into hash: "
        "templateId %d\n", templateId); /* debug */
#endif
        
    return TRUE;
}






/**
 * fbCollectorPostProcV9
 *
 * converts a buffer that was read as a netflow V9
 * into a buffer that "conforms" to IPFIX for the
 * rest of fixbuf to process it
 *
 * @param collector pointer to the collector state structure
 * @param dataBuf pointer to the netflow PDU
 * @param bufLen  the size from UDP of the message
 * @param err     glib error set when FALSE is returned with an
 *                informative message
 *
 * @return TRUE on success FALSE on error
 */
static gboolean     fbCollectorPostProcV9(
    fbCollector_t   *collector,
    uint8_t         *dataBuf,
    size_t          *bufLen,
    GError          **err)
{
    uint16_t                                recordCount;  
    uint16_t                                loop;         
    uint8_t                                 *msgOsetPtr = dataBuf;
    uint16_t                                *lengthCountPtr = NULL;
	uint32_t	                            netflowSeqNum;
    uint32_t                                ipfixRecordCount = 0;
    struct fbCollectorNetflowV9State_st     *transState = 
        (struct fbCollectorNetflowV9State_st *)collector->translatorState;
    uint32_t                                timeStamp;
    uint32_t                                obsDomain;
    uint16_t                                version;
    uint16_t                                amountRemoved = 0;
    uint32_t                                *seqNumPtr;

#if FB_NETFLOW_DEBUG == 1
    uint16_t                                dLoop; /* debug */


    /* debug */
    {
        uint8_t  *d_tPtr8 = (uint8_t *)dataBuf;
        uint16_t *d_tPtr = (uint16_t *)dataBuf;
        uint16_t d_setid = g_ntohs(*d_tPtr);
        d_tPtr++;
        uint16_t d_setLen = g_ntohs(*d_tPtr);
        fprintf(stderr, "\nversion: %d count: %d\n", 
            d_setid, d_setLen );
        for (dLoop=0; dLoop < *bufLen; dLoop++) {
            fprintf(stderr, "0x%02x ", (d_tPtr8)[dLoop]);
            if (0 == (dLoop+1)%4 && 0 != dLoop) fprintf(stderr, "\n");
        }
    }
#endif

    
    /* the buffer header has been partially converted into
    IPFIX, mostly meaning that the extra uptime has been
    dropped; now the count has to be converted into length,
    and all the different records need to be updated
    appropriately */
    READU16(msgOsetPtr, version);
    WRITEU16(msgOsetPtr, 0x0a);
    msgOsetPtr += sizeof(uint16_t);

    lengthCountPtr = (uint16_t *)msgOsetPtr;
    READU16INC(msgOsetPtr, recordCount);
    
    READU32INC(msgOsetPtr, timeStamp);
    
    seqNumPtr = (uint32_t *)msgOsetPtr;
    READU32INC(msgOsetPtr, netflowSeqNum);
    if (transState->netflowSeqNum != netflowSeqNum) {
        g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING,
            "NetFlow V9 sequence number mismatch, "
            "expecting %d received %d", transState->netflowSeqNum,
            netflowSeqNum);
         transState->netflowSeqNum = netflowSeqNum;
    }

    /* read the observation domain */
    READU32INC(msgOsetPtr, obsDomain);
    
    /* iterate through the flowsets */
    for (loop=0; loop < recordCount; loop++) {
        uint16_t setId;
        uint16_t recordLength;
        
        /* read the set ID, and adjust the reserved ones from 
        Netflow to IPFIX */
        READU16(msgOsetPtr, setId);
        if (0 == setId) {
            WRITEU16(msgOsetPtr, 2);
        } else if (1 == setId) {
            msgOsetPtr += sizeof(uint16_t);
            READU16INC(msgOsetPtr, recordLength);

            if (!netflowOptionsTemplateParse(collector, msgOsetPtr, 
                recordLength-4, err)) {
                    return FALSE;
            }

            /* life just got ugly, crunch this options section away and 
            warn that it got dumped
            
            +-- dataBuf
            |                   dataBuf+*bufLen --+  
            |                                     |
            v                                     v
            +-------------------------------------+
            |           |xx|   |                  |
            +-------------------------------------+
                        ^      ^
                        |      |
                        |      +- msgOsetPtr-4+recordLength
                        |
                        +-- msgOsetPtr-4
            
            need to remove the bad flow set type, everything from
            msgOsetPtr-4 to msgOsetPtr-4+recordLength, 
            the size of the part beyond that which has to move is:
            *bufLen - ( (msgOsetPtr-4+recordLength) - dataBuf )
            
            */
#if FB_NETFLOW_DEBUG == 1
            fprintf(stderr, "options removal code kicked in\n"); /*  debug */
#endif
            g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "NetFlow V9 section "
                "type %d is not handled, flow set is removed", setId);
            memmove(msgOsetPtr-4, msgOsetPtr+recordLength-4, 
                *bufLen - ((msgOsetPtr-4+recordLength) - dataBuf));
            amountRemoved += recordLength;
            msgOsetPtr -= 4;
            *bufLen -= recordLength;
            continue;
        } else if (setId < 256) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
                "NetFlow record type (%d) is not supported", setId);
            return FALSE;
        } 
        msgOsetPtr += sizeof(uint16_t);
        
        /* read the length of this record */
        READU16INC(msgOsetPtr, recordLength);

        /* need to do more with a template set than a data set */
        if (0 == setId) {
            /* template record */
            /* if the IPFIX standard info model was actually 
            compatible with the V9 info model as advertised, this
            piece of code would be sane
            */
            if (!netflowDataTemplateParse(collector, msgOsetPtr,
                (uint16_t *)(msgOsetPtr-sizeof(uint16_t)), dataBuf, 
                bufLen, err)) {
                return FALSE;
            }
            /* adjust the message pointer to skip over the payload of 
               the template record, read the size from the record 
               because it may have been updated in the template parse
               call (take into account that the pointer is 4-bytes
               into the record already (type & length)) */
            msgOsetPtr += 
                (g_ntohs(*(uint16_t *)(msgOsetPtr - sizeof(uint16_t))) -
                (2 * sizeof(uint16_t)));
        } else if (setId >= 128) {
            /* data record */
            struct fbCollectorNetflowV9TemplateHash_st *derTemplate = NULL;
            uint16_t numberRecordsInSet = 0;
            uint32_t bigSetId = (uint32_t) setId;
            derTemplate = g_hash_table_lookup(transState->templateHash, 
                (gconstpointer)bigSetId);
            if (NULL == derTemplate) {
                g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
                    "unknown template present, id %d", setId);
                return FALSE;
            }
            if (TRUE == derTemplate->optionsTemplate) {
                /* crunch this record! it is an options record, which
                we isn't supported */
                /*numberRecordsInSet = recordLength / derTemplate->templateLength;
                ipfixRecordCount += numberRecordsInSet; */
                memmove(msgOsetPtr-4, msgOsetPtr-4+recordLength, 
                    *bufLen - ((msgOsetPtr-4+recordLength) - dataBuf));
                amountRemoved += recordLength;
                msgOsetPtr -= 4;
                *bufLen -= recordLength;
#if FB_NETFLOW_DEBUG == 1
                fprintf(stderr, "options record removal "
                    "code kicked in\n"); /* debug */
#endif
            } else {
                numberRecordsInSet = recordLength / derTemplate->templateLength;
#if FB_NETFLOW_DEBUG == 1
                fprintf(stderr, "number of data records is %d (0x%x)\n", 
                    numberRecordsInSet, numberRecordsInSet); /* debug */
#endif
                ipfixRecordCount += numberRecordsInSet;
                msgOsetPtr += recordLength - 4;
            }
        } else {
            /* a record type we don't handle! (options record)*/
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
                "NetFlow record type is not understood, options record?");
            return FALSE;
        }
        
    }

    /* fixup the length value (from record count)*/
    *lengthCountPtr = g_htons(msgOsetPtr - dataBuf);

    
    /* increment the sequence number for the netflow side, which is 
    equal to the total number of records seen */
    /*transState->netflowSeqNum += recordCount;*/
    transState->netflowSeqNum++;

    /* fixup the sequence number */
    *seqNumPtr = g_htonl(transState->ipfixSeqNum);

    /* increment the ipfix record count with the number of relevent 
    records we observed*/
    transState->ipfixSeqNum += ipfixRecordCount;


#if FB_NETFLOW_DEBUG == 1
    /* debug */
    {
        uint8_t  *d_tPtr8 = (uint8_t *)dataBuf;
        uint16_t *d_tPtr = (uint16_t *)dataBuf;
        uint16_t d_setid = g_ntohs(*d_tPtr);
        d_tPtr++;
        uint16_t d_setLen = g_ntohs(*d_tPtr);
        fprintf(stderr, "\nversion: %d count: %d\n", 
            d_setid, d_setLen );
        for (dLoop=0; dLoop < *bufLen; dLoop++) {
            fprintf(stderr, "0x%02x ", (d_tPtr8)[dLoop]);
            if (0 == (dLoop+1)%4 && 0 != dLoop) fprintf(stderr, "\n");
        }
    }
#endif

    return TRUE;
}



/**
 * fbCollectorTransCloseV9
 *
 * frees the state included as part of the collector when the
 * Netflow V9 translator is enabled
 *
 * @param collector, pointer to the collector state structure
 *
 */
static void         fbCollectorTransCloseV9(
    fbCollector_t   *collector)
{
    struct fbCollectorNetflowV9State_st     *transState = 
        (struct fbCollectorNetflowV9State_st *)collector->translatorState;

    g_hash_table_destroy(transState->templateHash);
    transState->templateHash = NULL;

    if (NULL != collector->translatorState) {
        g_free(collector->translatorState);
    }

    return;
}

/**
 * templateHashDestroyHelper
 *
 * helps destroy the template hash by translating between the 
 * GLib GDestroyNotify function type definition and using the 
 * GLib slice free function
 *
 * @param datum pointer to the structure to be destroyed
 *
 */
static void         templateHashDestroyHelper (
    gpointer datum)
{
    g_slice_free(fbCollectorNetflowV9TemplateHash_t, datum);
}


static guint        fooHash (
    gconstpointer   key)
{
    return ((guint) key);
}

static gboolean     fooEqual (
    gconstpointer   alpha,
    gconstpointer   beta)
{
    guint   alphaInt = (guint)alpha;
    guint   betaInt = (guint)beta;
    
    if (alphaInt == betaInt) {
        return TRUE;
    }
    
    return FALSE;
}

/**
 *fbCollectorSetNetflowV9Translator
 *
 * this sets the collector input translator
 * to convert NetFlowV9 into IPFIX for the
 * given collector
 *
 * @param collector pointer to the collector state
 *        to perform Netflow V9 conversion on
 * @param err GError structure that holds the error
 *        message if an error occurs
 *
 *
 * @return TRUE on success, FALSE on error
 */
gboolean    fbCollectorSetNetflowV9Translator(
    fbCollector_t               *collector,
    GError                      **err)
{
    GHashTable *hashTable = NULL;
    struct fbCollectorNetflowV9State_st *nflowState = 
        g_malloc(sizeof(struct fbCollectorNetflowV9State_st));

    if (NULL == nflowState)
    {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_TRANSMISC,
            "failure to allocate Netflow V9 translator state");
        return FALSE;
    }

    hashTable = g_hash_table_new_full(fooHash, fooEqual, 
        NULL, templateHashDestroyHelper);

    if (NULL == hashTable) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_NETFLOWV9,
            "failure to allocate hash table for NetFlow session");
        return FALSE;
    }
#if FB_NETFLOW_DEBUG == 1
    fprintf(stderr, "Hash table address is %p for collector %p\n", 
        hashTable, collector); /* debug */
#endif

    nflowState->ipfixSeqNum = 0;
    nflowState->netflowSeqNum = 0;
    nflowState->templateHash = hashTable;

    return fbCollectorSetTranslator(collector, fbCollectorPostProcV9,
        fbCollectorDecodeV9MsgVL, fbCollectorMessageHeaderV9,
        fbCollectorTransCloseV9, nflowState, err);
    
}
