/*@internal
 *
 ** fbuf.c
 ** IPFIX Message buffer implementation
 **
 ** ------------------------------------------------------------------------
 ** Copyright (C) 2006-2008 Carnegie Mellon University. All Rights Reserved.
 ** ------------------------------------------------------------------------
 ** Authors: Brian Trammell <bht@cert.org>
 ** ------------------------------------------------------------------------
 ** GNU Lesser GPL Rights pursuant to Version 2.1, February 1999
 ** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
 ** ------------------------------------------------------------------------
 */

#define _FIXBUF_SOURCE_
#include <fixbuf/private.h>

#ident "$Id: fbuf.c 12302 2008-08-21 20:26:28Z inacio_svn $"

#define FB_MTU_MIN          32
#define FB_TCPLAN_NULL      -1

/* Debugger switches. We'll want to stick these in autoinc at some point. */
#define FB_DEBUG_TC         0
#define FB_DEBUG_TMPL       0
#define FB_DEBUG_WR         0
#define FB_DEBUG_RD         0
#define FB_DEBUG_LWR        0
#define FB_DEBUG_LRD        0

typedef struct fbTranscodePlan_st {
    fbTemplate_t    *s_tmpl;
    fbTemplate_t    *d_tmpl;
    int32_t         *si;
} fbTranscodePlan_t;

struct fBuf_st {
    /** Transport session. Contains template and sequence number state. */
    fbSession_t         *session;
    /** Exporter. Writes messages to a remote endpoint on flush. */
    fbExporter_t        *exporter;
    /** Collector. Reads messages from a remote endpoint on demand. */
    fbCollector_t       *collector;
    /** Automatic mode flag */
    gboolean            automatic;
    /** Cached transcoder plan */
    fbTranscodePlan_t   *tcplan;
    /** Current internal template ID. */
    uint16_t            int_tid;
    /** Current external template ID. */
    uint16_t            ext_tid;
    /** Current special set ID. */
    uint16_t            spec_tid;
    /** Current internal template. */
    fbTemplate_t        *int_tmpl;
    /** Current external template. */
    fbTemplate_t        *ext_tmpl;
    /** Export time in seconds since 0UTC 1 Jan 1970 */
    uint32_t            extime;
    /** Record counter. */
    uint32_t            rc;
    /**
     * Current position pointer.
     * Pointer to the next byte in the buffer to be written or read.
     */
    uint8_t             *cp;
    /**
     * Pointer to first byte in the buffer in the current message.
     * NULL if there is no current message.
     */
    uint8_t             *msgbase;
    /**
     * Message end position pointer.
     * Pointer to first byte in the buffer after the current message.
     */
    uint8_t             *mep;
    /**
     * Pointer to first byte in the buffer in the current message.
     * NULL if there is no current message.
     */
    uint8_t             *setbase;
    /**
     * Set end position pointer.
     * Valid only after a call to fBufNextSetHeader() (called by fBufNext()).
     */
    uint8_t             *sep;
    /** Message buffer. */
    uint8_t             buf[FB_MSGLEN_MAX+1];
};

/*==================================================================
 *
 * Debugger Functions
 *
 *==================================================================*/

#if FB_DEBUG_WR || FB_DEBUG_RD || FB_DEBUG_TC

static uint32_t fBufDebugHexLine(
    GString             *str,
    const char          *lpfx,
    uint8_t             *cp,
    uint32_t            lineoff,
    uint32_t            buflen)
{
    uint32_t            cwr = 0, twr = 0;

    /* stubbornly refuse to print nothing */
    if (!buflen) return 0;

    /* print line header */
    g_string_append_printf(str, "%s %04x:", lpfx, lineoff);

    /* print hex characters */
    for (twr = 0; twr < 16; twr++) {
        if (buflen) {
            g_string_append_printf(str, " %02hhx", cp[twr]);
            cwr++; buflen--;
        } else {
            g_string_append(str, "   ");
        }
    }

    /* print characters */
    g_string_append_c(str, ' ');
    for (twr = 0; twr < cwr; twr++) {
        if (cp[twr] > 32 && cp[twr] < 128) {
            g_string_append_c(str, cp[twr]);
        } else {
            g_string_append_c(str, '.');
        }
    }
    g_string_append_c(str, '\n');

    return cwr;
}

static void fBufDebugHex(
    const char          *lpfx,
    uint8_t             *buf,
    uint32_t            len)
{
    GString             *str = g_string_new("");
    uint32_t            cwr = 0, lineoff = 0;

    do {
        cwr = fBufDebugHexLine(str, lpfx, buf, lineoff, len);
        buf += cwr; len -= cwr; lineoff += cwr;
    } while (cwr == 16);

    fprintf(stderr,"%s", str->str);
    g_string_free(str, TRUE);
}

#endif

#if FB_DEBUG_WR || FB_DEBUG_RD

#if FB_DEBUG_TC

static void fBufDebugTranscodePlan(
    fbTranscodePlan_t   *tcplan)
{
    int                 i;

    fprintf(stderr, "transcode plan %p -> %p\n",
            tcplan->s_tmpl, tcplan->d_tmpl);
    for (i = 0; i < tcplan->d_tmpl->ie_count; i++) {
        fprintf(stderr, "\td[%2u]=s[%2d]\n", i, tcplan->si[i]);
    }
}

static void fBufDebugTranscodeOffsets(
    fbTemplate_t        *tmpl,
    uint16_t            *offsets)
{
    int                 i;

    fprintf(stderr, "offsets %p\n", tmpl);
    for (i = 0; i < tmpl->ie_count; i++) {
        fprintf(stderr, "\to[%2u]=%4x\n", i, offsets[i]);
    }
}

#endif

static void fBufDebugBuffer(
    const char      *label,
    fBuf_t          *fbuf,
    size_t          len,
    gboolean        reverse)
{
    uint8_t         *xcp = fbuf->cp - len;
    uint8_t         *rcp = reverse ? xcp : fbuf->cp;

    fprintf(stderr, "%s len %5lu mp %5u (0x%04x) sp %5u mr %5u sr %5u\n",
            label, len, rcp - fbuf->msgbase, rcp - fbuf->msgbase,
            fbuf->setbase ? (rcp - fbuf->setbase) : 0,
            fbuf->mep - fbuf->cp,
            fbuf->sep ? (fbuf->sep - fbuf->cp) : 0);

    fBufDebugHex(label, rcp, len);
}

#endif

/*==================================================================
 *
 * Transcoder Functions
 *
 *==================================================================*/

#define FB_TC_SBC_OFF(_need_)                                           \
    if (s_rem < (_need_)) {                                             \
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOM,                 \
                    "End of message. "                                  \
                    "Underrun on transcode offset calculation "         \
                    "(need %lu bytes, %lu available)",                  \
                    (unsigned long)(_need_), (unsigned long)s_rem);     \
        goto err;                                                       \
    }

#define FB_TC_DBC(_need_, _op_)                                             \
    if (*d_rem < (_need_)) {                                                \
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOM,                     \
                    "End of message. "                                      \
                    "Overrun on %s (need %lu bytes, %lu available)",        \
                    (_op_), (unsigned long)(_need_), (unsigned long)*d_rem);\
        return FALSE;                                                       \
    }

/**
 * fbTranscodePlan
 *
 *
 *
 */
static fbTranscodePlan_t *fbTranscodePlan(
    fBuf_t                  *fbuf,
    fbTemplate_t            *s_tmpl,
    fbTemplate_t            *d_tmpl)
{
    void                    *sik, *siv;
    uint32_t                i;

    /* check to see if plan is cached */
    if (fbuf->tcplan &&
        fbuf->tcplan->s_tmpl == s_tmpl &&
        fbuf->tcplan->d_tmpl == d_tmpl) {
        return fbuf->tcplan;
    }

    /* free existing plan */
    if (fbuf->tcplan) {
        g_free(fbuf->tcplan->si);
        g_slice_free(fbTranscodePlan_t, fbuf->tcplan);
    }

    /* create new transcode plan and cache it */
    fbuf->tcplan = g_slice_new0(fbTranscodePlan_t);

    /* fill in template refs */
    fbuf->tcplan->s_tmpl = s_tmpl;
    fbuf->tcplan->d_tmpl = d_tmpl;
    fbuf->tcplan->si = g_new0(int32_t, d_tmpl->ie_count);

    /* for each destination element */
    for (i = 0; i < d_tmpl->ie_count; i++) {
        /* find source index */
        if (g_hash_table_lookup_extended(s_tmpl->indices,
                                         d_tmpl->ie_ary[i],
                                         &sik, &siv)) {
            fbuf->tcplan->si[i] = GPOINTER_TO_INT(siv);
        } else {
            fbuf->tcplan->si[i] = FB_TCPLAN_NULL;
        }
    }

    return fbuf->tcplan;
}

/**
 * fbTranscodeFreeVarlenOffsets
 *
 *
 *
 *
 */
static void         fbTranscodeFreeVarlenOffsets(
    fbTemplate_t        *s_tmpl,
    uint16_t            *offsets)
{
    if (s_tmpl->is_varlen) g_free(offsets);
}

/**
 *
 * Macros for decode reading
 */


#if HAVE_ALIGNED_ACCESS_REQUIRED

#define FB_READ_U16(_val_, _ptr_) {                     \
    uint16_t _x;                                        \
    memcpy(&_x, _ptr_, sizeof(uint16_t));               \
    _val_ = g_ntohs(_x);                                \
}

#define FB_READ_U32(_val_, _ptr_) {                     \
    uint32_t _x;                                        \
    memcpy(&_x, _ptr_, sizeof(uint32_t));               \
    _val_ = g_ntohl(_x);                                \
}

#else

#define FB_READ_U16(_val_,_ptr_) {                      \
    _val_ = g_ntohs(*_ptr_);                            \
}

#define FB_READ_U32(_val_,_ptr_) {                      \
    _val_ = g_ntohl(*_ptr_);                            \
}

#endif


/**
 * fbTranscodeOffsets
 *
 *
 *
 */
static ssize_t      fbTranscodeOffsets(
    fbTemplate_t        *s_tmpl,
    uint8_t             *s_base,
    uint32_t            s_rem,
    gboolean            decode,
    uint16_t            **offsets_out,
    GError              **err)
{
    fbInfoElement_t     *s_ie;
    uint8_t             *sp;
    uint16_t            *offsets;
    uint32_t            s_len, i;

    /* short circuit - return offset cache if present in template */
    if (s_tmpl->off_cache) {
        if (offsets_out) *offsets_out = s_tmpl->off_cache;
        return s_tmpl->off_cache[s_tmpl->ie_count];
    }

    /* create new offsets array */
    offsets = g_new0(uint16_t, s_tmpl->ie_count + 1);

    /* populate it */
    for (i = 0, sp = s_base; i < s_tmpl->ie_count; i++) {
        offsets[i] = sp - s_base;
        s_ie = s_tmpl->ie_ary[i];
        if (s_ie->len == FB_IE_VARLEN) {
            if (decode) {
                FB_TC_SBC_OFF(1);
                s_len = *sp;
                if (s_len < 255) {
                    sp += 1; s_rem -= 1;
                } else {
                    FB_TC_SBC_OFF(3);
                    FB_READ_U16(s_len,(sp+1));
                    sp += 3; s_rem -= 3;
                }
                FB_TC_SBC_OFF(s_len);
                sp += s_len; s_rem -= s_len;
            } else {
                FB_TC_SBC_OFF(sizeof(fbVarfield_t));
                sp += sizeof(fbVarfield_t); s_rem -= sizeof(fbVarfield_t);
            }
        } else {
            FB_TC_SBC_OFF(s_ie->len);
            sp += s_ie->len; s_rem -= s_ie->len;
        }
    }

    /* get EOR offset */
    s_len = offsets[i] = sp - s_base;

    /* cache offsets if possible */
    if (!s_tmpl->is_varlen && offsets_out) {
        s_tmpl->off_cache = offsets;
    }

    /* return offsets if possible */
    if (offsets_out) {
        *offsets_out = offsets;
    } else {
        *offsets_out = NULL;
        fbTranscodeFreeVarlenOffsets(s_tmpl, offsets);
    }

    /* return EOR offset */
    return s_len;

  err:
    g_free(offsets);
    return -1;
}


/**
 * fbTranscodeZero
 *
 *
 *
 *
 *
 */
static gboolean fbTranscodeZero(
    uint8_t             **dp,
    uint32_t            *d_rem,
    uint32_t            len,
    GError              **err)
{
    /* Check for write overrun */
    FB_TC_DBC(len, "zero transcode");

    /* fill zeroes */
    memset(*dp, 0, len);

    /* maintain counters */
    *dp += len; *d_rem -= len;

    return TRUE;
}



/**
 *  fbTranscodeSwap
 *
 *
 *
 *
 *
 */
static void fbTranscodeSwap(
    uint8_t             *a,
    uint32_t            len)
{
    uint32_t            i;
    uint8_t             t;

    for (i = 0; i < len/2; i++) {
        t = a[i];
        a[i] = a[(len-1)-i];
        a[(len-1)-i] = t;
    }
}

#if G_BYTE_ORDER == G_BIG_ENDIAN



/**
 * fbTranscodeFixedBigEndian
 *
 *
 *
 *
 *
 */
static gboolean fbTranscodeFixedBigEndian(
    uint8_t             *sp,
    uint8_t             **dp,
    uint32_t            *d_rem,
    uint32_t            s_len,
    uint32_t            d_len,
    uint32_t            flags,
    GError              **err)
{
    FB_TC_DBC(d_len, "fixed transcode");

    if (s_len == d_len) {
        memcpy(*dp, sp, d_len);
    } else if (s_len > d_len) {
        if (flags & FB_IE_F_ENDIAN) {
            memcpy(*dp, sp + (s_len - d_len), d_len);
        } else {
            memcpy(*dp, sp, d_len);
        }
    } else {
        memset(*dp, 0, d_len);
        if (flags & FB_IE_F_ENDIAN) {
            memcpy(*dp + (d_len - s_len), sp, s_len);
        } else {
            memcpy(*dp, sp, s_len);
        }
    }

    /* maintain counters */
    *dp += d_len; *d_rem -= d_len;

    return TRUE;
}

#define fbEncodeFixed fbTranscodeFixedBigEndian
#define fbDecodeFixed fbTranscodeFixedBigEndian
#else


/**
 * fbEncodeFixedLittleEndian
 *
 *
 *
 *
 *
 */
static gboolean fbEncodeFixedLittleEndian(
    uint8_t             *sp,
    uint8_t             **dp,
    uint32_t            *d_rem,
    uint32_t            s_len,
    uint32_t            d_len,
    uint32_t            flags,
    GError              **err)
{
    FB_TC_DBC(d_len, "fixed LE encode");

    if (s_len == d_len) {
        memcpy(*dp, sp, d_len);
    } else if (s_len > d_len) {
        if (flags & FB_IE_F_ENDIAN) {
            memcpy(*dp, sp, d_len);
        } else {
            memcpy(*dp, sp + (s_len - d_len), d_len);
        }
    } else {
        memset(*dp, 0, d_len);
        if (flags & FB_IE_F_ENDIAN) {
            memcpy(*dp, sp, s_len);
        } else {
            memcpy(*dp + (d_len - s_len), sp, s_len);
        }
    }

    /* swap bytes at destination if necessary */
    if (d_len > 1 && (flags & FB_IE_F_ENDIAN)) {
        fbTranscodeSwap(*dp, d_len);
    }

    /* maintain counters */
    *dp += d_len; *d_rem -= d_len;

    return TRUE;
}


/**
 * fbDecodeFixedLittleEndian
 *
 *
 *
 *
 *
 */
static gboolean fbDecodeFixedLittleEndian(
    uint8_t             *sp,
    uint8_t             **dp,
    uint32_t            *d_rem,
    uint32_t            s_len,
    uint32_t            d_len,
    uint32_t            flags,
    GError              **err)
{
    FB_TC_DBC(d_len, "fixed LE decode");

    if (s_len == d_len) {
        memcpy(*dp, sp, d_len);
    } else if (s_len > d_len) {
        if (flags & FB_IE_F_ENDIAN) {
            memcpy(*dp, sp + (s_len - d_len), d_len);
        } else {
            memcpy(*dp, sp, d_len);
        }
    } else {
        memset(*dp, 0, d_len);
        if (flags & FB_IE_F_ENDIAN) {
            memcpy(*dp + (d_len - s_len), sp, s_len);
        } else {
            memcpy(*dp, sp, s_len);
        }
    }

    /* swap bytes at destination if necessary */
    if (d_len > 1 && (flags & FB_IE_F_ENDIAN)) {
        fbTranscodeSwap(*dp, d_len);
    }

    /* maintain counters */
    *dp += d_len; *d_rem -= d_len;

    return TRUE;
}

#define fbEncodeFixed fbEncodeFixedLittleEndian
#define fbDecodeFixed fbDecodeFixedLittleEndian
#endif


/**
 * fbEncodeVarfield
 *
 *
 *
 *
 *
 */
static gboolean fbEncodeVarfield(
    uint8_t             *sp,
    uint8_t             **dp,
    uint32_t            *d_rem,
    uint32_t            flags,
    GError              **err)
{
    uint32_t            d_len;
    uint16_t            sll;
    fbVarfield_t         *sv = (fbVarfield_t *)sp;

    /* calculate total destination length */
    d_len = sv->len + ((sv->len < 255) ? 1 : 3);

    /* Check buffer bounds */
    FB_TC_DBC(d_len, "variable-length encode");

    /* emit IPFIX variable length */
    if (sv->len < 255) {
        **dp = (uint8_t)sv->len;
        *dp += 1;
    } else {
        **dp = 255;
        sll = g_htons((uint16_t)sv->len);
        memcpy(*dp + 1, &sll, sizeof(uint16_t));
        *dp += 3;
    }

    /* emit buffer contents */
    if (sv->len && sv->buf) memcpy(*dp, sv->buf, sv->len);

    /* maintain counters */
    *dp += sv->len; *d_rem -= d_len;

    return TRUE;
}


/**
 * fbDecodeVarfield
 *
 *
 *
 *
 *
 */
static gboolean fbDecodeVarfield(
    uint8_t             *sp,
    uint8_t             **dp,
    uint32_t            *d_rem,
    uint32_t            flags,
    GError              **err)
{
    uint16_t            s_len;
    fbVarfield_t         *dv = (fbVarfield_t *)*dp;

    /* calculate total source length */
    s_len = *sp;
    if (s_len < 255) {
        sp += 1;
    } else {
        memcpy(&s_len, sp + 1, sizeof(uint16_t));
        s_len = g_ntohs(s_len);
        sp += 3;
    }

    /* Okay. We know how long the source is. Check buffer bounds. */
    FB_TC_DBC(sizeof(fbVarfield_t), "variable-length decode");

    /* Do transcode. Don't copy; fbVarfield_t's semantics allow us just
       to return a pointer into the read buffer. */
    dv->len = (uint32_t)s_len;
    dv->buf = s_len ? sp : NULL;

    /* maintain counters */
    *dp += sizeof(fbVarfield_t); *d_rem -= sizeof(fbVarfield_t);

    return TRUE;
}


/**
 * fbTranscode
 *
 *
 *
 *
 *
 */
static gboolean fbTranscode(
    fBuf_t              *fbuf,
    gboolean            decode,
    uint8_t             *s_base,
    uint8_t             *d_base,
    size_t              *s_len,
    size_t              *d_len,
    GError              **err)
{
    fbTranscodePlan_t   *tcplan;
    fbTemplate_t        *s_tmpl, *d_tmpl;
    ssize_t             s_len_offset;
    uint16_t            *offsets;
    uint8_t             *dp;
    uint32_t            s_off, d_rem, i;
    fbInfoElement_t     *s_ie, *d_ie;
    gboolean            ok = TRUE;

    /* initialize walk of dest buffer */
    dp = d_base; d_rem = *d_len;

    /* select templates for transcode */
    if (decode) {
        s_tmpl = fbuf->ext_tmpl;
        d_tmpl = fbuf->int_tmpl;
    } else {
        s_tmpl = fbuf->int_tmpl;
        d_tmpl = fbuf->ext_tmpl;
    }

    /* get a transcode plan */
    tcplan = fbTranscodePlan(fbuf, s_tmpl, d_tmpl);

    /* get source record length and offsets */
    if ((s_len_offset = fbTranscodeOffsets(s_tmpl, s_base, *s_len,
                                           decode, &offsets, err)) < 0)
        return FALSE;
    *s_len = s_len_offset;

#if FB_DEBUG_TC && FB_DEBUG_RD && FB_DEBUG_WR
    fBufDebugTranscodePlan(tcplan);
    if (offsets) fBufDebugTranscodeOffsets(s_tmpl, offsets);
    fBufDebugHex("tsrc", s_base, *s_len);
#elif FB_DEBUG_TC && FB_DEBUG_RD
    if (decode) {
        fBufDebugTranscodePlan(tcplan);
        if (offsets) fBufDebugTranscodeOffsets(s_tmpl, offsets);
        fBufDebugHex("tsrc", s_base, *s_len);
    }
#elif FB_DEBUG_TC && FB_DEBUG_WR
    if (!decode) {
        fBufDebugTranscodePlan(tcplan);
        if (offsets) fBufDebugTranscodeOffsets(s_tmpl, offsets);
        fBufDebugHex("tsrc", s_base, *s_len);
    }
#endif

    /* iterate over destination IEs, copying from source */
    for (i = 0; i < d_tmpl->ie_count; i++) {
        /* Get pointers to information elements and source offset */
        d_ie = d_tmpl->ie_ary[i];
        s_ie = (tcplan->si[i] == FB_TCPLAN_NULL) ?
            NULL : s_tmpl->ie_ary[tcplan->si[i]];
        s_off = s_ie ? offsets[tcplan->si[i]] : 0;

        if (s_ie == NULL) {
            /* Null source */
            uint32_t null_len =
                ((d_ie->len == FB_IE_VARLEN)
                    ? ( decode ? sizeof(fbVarfield_t) : 1 )
                    : d_ie->len);
            if (!(ok = fbTranscodeZero(&dp, &d_rem, null_len, err))) {
                goto end;
            }
        } else if (s_ie->len != FB_IE_VARLEN && d_ie->len != FB_IE_VARLEN) {
            /* Fixed transcode */
            if (decode) {
                ok = fbDecodeFixed(s_base + s_off, &dp, &d_rem,
                                   s_ie->len, d_ie->len,
                                   d_ie->flags, err);
            } else {
                ok = fbEncodeFixed(s_base + s_off, &dp, &d_rem,
                                   s_ie->len, d_ie->len,
                                   d_ie->flags, err);
            }
            if (!ok) {
                goto end;
            }
        } else if (s_ie->len == FB_IE_VARLEN && d_ie->len == FB_IE_VARLEN) {
            /* Varlen transcode */
            if (decode) {
                ok = fbDecodeVarfield(s_base + s_off, &dp, &d_rem,
                                      d_ie->flags, err);
            } else {
                ok = fbEncodeVarfield(s_base + s_off, &dp, &d_rem,
                                      d_ie->flags, err);
            }
            if (!ok) {
                goto end;
            }
        } else {
            /* Fixed to varlen or vice versa */
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IMPL,
                        "Transcoding between fixed and varlen IE "
                        "not supported by this version of libfixbuf.");
            ok = FALSE;
            goto end;
        }
    }

    /* Return destination length */
    *d_len = dp - d_base;

#if FB_DEBUG_TC && FB_DEBUG_RD && FB_DEBUG_WR
    fBufDebugHex("tdst", d_base, *d_len);
#elif FB_DEBUG_TC && FB_DEBUG_RD
    if (decode) fBufDebugHex("tdst", d_base, *d_len);
#elif FB_DEBUG_TC && FB_DEBUG_WR
    if (!decode) fBufDebugHex("tdst", d_base, *d_len);
#endif

    /* All done */
  end:
    fbTranscodeFreeVarlenOffsets(s_tmpl, offsets);
    return ok;
}

/*==================================================================
 *
 * Common Buffer Management Functions
 *
 *==================================================================*/

#define FB_REM_MSG(_fbuf_) (_fbuf_->mep - _fbuf_->cp)

#define FB_REM_SET(_fbuf_) (_fbuf_->sep - _fbuf_->cp)


/**
 * fBufRewind
 *
 *
 *
 *
 *
 */
void            fBufRewind(
    fBuf_t          *fbuf)
{
    /* Reset the buffer */
    fbuf->cp = fbuf->buf;
    fbuf->mep = fbuf->cp;

    /* No message or set headers in buffer */
    fbuf->msgbase = NULL;
    fbuf->setbase = NULL;
    fbuf->sep = NULL;

    /* No records in buffer either */
    fbuf->rc = 0;
}


/**
 * fBufSetInternalTemplate
 *
 *
 *
 *
 *
 */
gboolean        fBufSetInternalTemplate(
    fBuf_t          *fbuf,
    uint16_t        int_tid,
    GError          **err)
{
    /* Look up new internal template if necessary */
    if (!fbuf->int_tmpl || fbuf->int_tid != int_tid) {
        fbuf->int_tid = int_tid;
        fbuf->int_tmpl = fbSessionGetTemplate(fbuf->session, TRUE, int_tid, err);
        if (!fbuf->int_tmpl) return FALSE;
    }

#if FB_DEBUG_TMPL
    fbTemplateDebug("int", int_tid, fbuf->int_tmpl);
#endif
    return TRUE;
}


/**
 * fBufSetAutomaticMode
 *
 *
 *
 *
 *
 */
void            fBufSetAutomaticMode(
    fBuf_t          *fbuf,
    gboolean        automatic)
{
    fbuf->automatic = automatic;
}


/**
 * fBufGetSession
 *
 *
 *
 *
 *
 */
fbSession_t     *fBufGetSession(
    fBuf_t          *fbuf)
{
    return fbuf->session;
}


/**
 * fBufFree
 *
 *
 *
 *
 *
 */
void            fBufFree(
    fBuf_t          *fbuf)
{
    if (fbuf->exporter) {
        fbExporterFree(fbuf->exporter);
    }

    if (fbuf->collector) {
        fbCollectorFree(fbuf->collector);
    }

    fbSessionFree(fbuf->session);

    g_slice_free(fBuf_t, fbuf);
}

/*==================================================================
 *
 * Buffer Append (Writer) Functions
 *
 *==================================================================*/

#if HAVE_ALIGNED_ACCESS_REQUIRED

#define FB_APPEND_U16(_val_) {                          \
    uint16_t _x = g_htons(_val_);                       \
    memcpy(fbuf->cp, &(_x), sizeof(uint16_t));          \
    fbuf->cp += sizeof(uint16_t);                       \
}

#define FB_APPEND_U32(_val_) {                          \
    uint32_t _x = g_htonl(_val_);                       \
    memcpy(fbuf->cp, &(_x), sizeof(uint32_t));          \
    fbuf->cp += sizeof(uint32_t);                       \
}

#else

#define FB_APPEND_U16(_val_)                            \
    *(uint16_t *)fbuf->cp = g_htons(_val_);             \
    fbuf->cp += sizeof(uint16_t);

#define FB_APPEND_U32(_val_)                            \
    *(uint32_t *)fbuf->cp = g_htonl(_val_);             \
    fbuf->cp += sizeof(uint32_t);

#endif


/**
 * fBufAppendMessageHeader
 *
 *
 *
 *
 *
 */
static void     fBufAppendMessageHeader(
    fBuf_t          *fbuf)
{
    /* can only append message header at start of buffer */
    g_assert(fbuf->cp == fbuf->buf);

    /* can only append message header if we have an exporter */
    g_assert(fbuf->exporter);

    /* get MTU from exporter */
    fbuf->mep += fbExporterGetMTU(fbuf->exporter);
    g_assert(FB_REM_MSG(fbuf) > FB_MTU_MIN);

    /* set message base pointer to show we have an active message */
    fbuf->msgbase = fbuf->cp;

    /* add version to buffer */
    FB_APPEND_U16(0x000A);

    /* add message length to buffer */
    FB_APPEND_U16(0);

    /* add export time to buffer */
    if (fbuf->extime) {
        FB_APPEND_U32(fbuf->extime);
    } else {
        FB_APPEND_U32(time(NULL));
    }

    /* add sequence number to buffer */
    FB_APPEND_U32(fbSessionGetSequence(fbuf->session));

    /* add observation domain ID to buffer */
    FB_APPEND_U32(fbSessionGetDomain(fbuf->session));

#if FB_DEBUG_WR
    fBufDebugBuffer("amsg", fbuf, 16, TRUE);
#endif

}


/**
 * fBufAppendSetHeader
 *
 *
 *
 *
 *
 */
static gboolean fBufAppendSetHeader(
    fBuf_t          *fbuf,
    GError          **err)
{
    uint16_t        set_id, set_minlen;

    /* Select set ID and minimum set size based on special TID */
    if (fbuf->spec_tid) {
        set_id = fbuf->spec_tid;
        set_minlen = 4;
    } else {
        set_id = fbuf->ext_tid;
        set_minlen = (fbuf->ext_tmpl->ie_len + 4);
    }

    /* Need enough space in the message for a set header and a record */
    if (FB_REM_MSG(fbuf) < set_minlen) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOM,
                    "End of message. "
                    "Overrun on set header append "
                    "(need %u bytes, %u available)",
                    set_minlen, FB_REM_MSG(fbuf));
        return FALSE;
    }

    /* set set base pointer to show we have an active set */
    fbuf->setbase = fbuf->cp;

    /* add set ID to buffer */
    FB_APPEND_U16(set_id);

    /* add set length to buffer */
    FB_APPEND_U16(0);

#if FB_DEBUG_WR
    fBufDebugBuffer("aset", fbuf, 4, TRUE);
#endif

    return TRUE;
}


/**
 * fBufAppendSetClose
 *
 *
 *
 *
 *
 */
static void fBufAppendSetClose(
    fBuf_t          *fbuf)
{
    uint16_t        setlen;

    /* If there's an active set... */
    if (fbuf->setbase) {
        /* store set length */
        setlen = g_htons(fbuf->cp - fbuf->setbase);
        memcpy(fbuf->setbase + 2, &setlen, sizeof(setlen));

#if FB_DEBUG_WR
    fBufDebugHex("cset", fbuf->setbase, 4);
#endif

        /* deactivate set */
        fbuf->setbase = NULL;
    }
}


/**
 * fBufSetExportTemplate
 *
 *
 *
 *
 *
 */
gboolean        fBufSetExportTemplate(
    fBuf_t          *fbuf,
    uint16_t        ext_tid,
    GError          **err)
{
    /* Look up new external template if necessary */
    if (!fbuf->ext_tmpl || fbuf->ext_tid != ext_tid) {
        fbuf->ext_tid = ext_tid;
        fbuf->ext_tmpl = fbSessionGetTemplate(fbuf->session, FALSE, ext_tid, err);
        if (!fbuf->ext_tmpl) return FALSE;

        /* Change of template means new set */
        fBufAppendSetClose(fbuf);
    }

#if FB_DEBUG_TMPL
    fbTemplateDebug("ext", ext_tid, fbuf->ext_tmpl);
#endif

    /* If we're here we're done. */
    return TRUE;
}


/**
 * fBufAppendTemplateSingle
 *
 *
 *
 *
 *
 */
static gboolean fBufAppendTemplateSingle(
    fBuf_t          *fbuf,
    uint16_t        tmpl_id,
    fbTemplate_t    *tmpl,
    gboolean        revoke,
    GError          **err)
{
    uint16_t        spec_tid, tmpl_len, ie_count, scope_count;
    int             i;

    /* Force message closed to start a new template message */
    if (!fbuf->spec_tid) {
        fbuf->spec_tid = (tmpl->scope_count) ? FB_TID_OTS : FB_TID_TS;
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOM,
                    "End of message. "
                    "Must start new message for template export.");
        return FALSE;
    }

    /* Start a new message if necessary */
    if (!fbuf->msgbase) {
        fBufAppendMessageHeader(fbuf);
    }

    /* Check for set ID change */
    spec_tid = (tmpl->scope_count) ? FB_TID_OTS : FB_TID_TS;
    if (fbuf->spec_tid != spec_tid) {
        fbuf->spec_tid = spec_tid;
        fBufAppendSetClose(fbuf);
    }

    /* Start a new set if necessary */
    if (!fbuf->setbase) {
        if (!fBufAppendSetHeader(fbuf, err)) return FALSE;
    }

    /*
     * Calculate template length and IE count based on whether this
     * is a revocation.
     */
    if (revoke) {
        tmpl_len = 4;
        ie_count = 0;
        scope_count = 0;
    } else {
        tmpl_len = tmpl->tmpl_len;
        ie_count = tmpl->ie_count;
        scope_count = tmpl->scope_count;
    }

    /* Ensure we have enough space for the template in the message */
    if (FB_REM_MSG(fbuf) < tmpl_len) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOM,
                    "End of message. "
                    "Overrun on template append "
                    "(need %u bytes, %u available)",
                    tmpl_len, FB_REM_MSG(fbuf));
        return FALSE;
    }

    /* Copy the template header to the message */
    FB_APPEND_U16(tmpl_id);
    FB_APPEND_U16(ie_count);

    /* Copy scope IE count if present */
    if (scope_count) {
        FB_APPEND_U16(scope_count);
    }

    /* Now copy information element specifiers to the buffer */
    for (i = 0; i < ie_count; i++) {
        if (tmpl->ie_ary[i]->ent) {
            FB_APPEND_U16(IPFIX_ENTERPRISE_BIT | tmpl->ie_ary[i]->num);
            FB_APPEND_U16(tmpl->ie_ary[i]->len);
            FB_APPEND_U32(tmpl->ie_ary[i]->ent);
        } else {
            FB_APPEND_U16(tmpl->ie_ary[i]->num);
            FB_APPEND_U16(tmpl->ie_ary[i]->len);
        }
    }

    /* Template records are records too. Increment record count. */
    /* Actually, no they're not. Odd. */
    /* ++(fbuf->rc); */

#if FB_DEBUG_TMPL
    fbTemplateDebug("apd", tmpl_id, tmpl);
#endif

#if FB_DEBUG_WR
    fBufDebugBuffer("atpl", fbuf, tmpl_len, TRUE);
#endif

    /* Done */
    return TRUE;
}


/**
 * fBufAppendTemplate
 *
 *
 *
 *
 *
 */
gboolean        fBufAppendTemplate(
    fBuf_t          *fbuf,
    uint16_t        tmpl_id,
    fbTemplate_t    *tmpl,
    gboolean        revoke,
    GError          **err)
{
    /* Attempt single append */
    if (fBufAppendTemplateSingle(fbuf, tmpl_id, tmpl, revoke, err)) return TRUE;

    /* Fail if not EOM or not automatic */
    if (!g_error_matches(*err, FB_ERROR_DOMAIN, FB_ERROR_EOM) ||
        !fbuf->automatic) return FALSE;

    /* Retryable. Clear error. */
    g_clear_error(err);

    /* Emit message */
    if (!fBufEmit(fbuf, err)) return FALSE;

    /* Retry single append */
    return fBufAppendTemplateSingle(fbuf, tmpl_id, tmpl, revoke, err);
}


/**
 * fBufAppendSingle
 *
 *
 *
 *
 *
 */
static gboolean fBufAppendSingle(
    fBuf_t          *fbuf,
    uint8_t         *recbase,
    size_t          recsize,
    GError          **err)
{
    size_t          bufsize;

    /* Buffer must have active templates */
    g_assert(fbuf->int_tmpl);
    g_assert(fbuf->ext_tmpl);

    /* Force message closed to finish any active template message */
    if (fbuf->spec_tid) {
        fbuf->spec_tid = 0;
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOM,
                    "End of message. "
                    "Must start new message after template export.");
        return FALSE;
    }

    /* Start a new message if necessary */
    if (!fbuf->msgbase) {
        fBufAppendMessageHeader(fbuf);
    }

    /* Cancel special set mode if necessary */
    if (fbuf->spec_tid) {
        fbuf->spec_tid = 0;
        fBufAppendSetClose(fbuf);
    }

    /* Start a new set if necessary */
    if (!fbuf->setbase) {
        if (!fBufAppendSetHeader(fbuf, err))
            return FALSE;
    }

    /* Transcode bytes into buffer */
    bufsize = FB_REM_MSG(fbuf);
    if (!fbTranscode(fbuf, FALSE, recbase, fbuf->cp, &recsize, &bufsize, err))
        return FALSE;

    /* Move current pointer forward by number of bytes written */
    fbuf->cp += bufsize;

    /* Increment record count */
    ++(fbuf->rc);

#if FB_DEBUG_WR
    fBufDebugBuffer("arec", fbuf, bufsize, TRUE);
#endif

    /* Done */
    return TRUE;
}


/**
 * fBufAppend
 *
 *
 *
 *
 *
 */
gboolean        fBufAppend(
    fBuf_t          *fbuf,
    uint8_t         *recbase,
    size_t          recsize,
    GError          **err)
{
    /* Attempt single append */
    if (fBufAppendSingle(fbuf, recbase, recsize, err)) return TRUE;

    /* Fail if not EOM or not automatic */
    if (!g_error_matches(*err, FB_ERROR_DOMAIN, FB_ERROR_EOM) ||
        !fbuf->automatic) return FALSE;

    /* Retryable. Clear error. */
    g_clear_error(err);

    /* Emit message */
    if (!fBufEmit(fbuf, err)) return FALSE;

    /* Retry single append */
    return fBufAppendSingle(fbuf, recbase, recsize, err);
}


/**
 * fBufEmit
 *
 *
 *
 *
 *
 */
gboolean        fBufEmit(
    fBuf_t          *fbuf,
    GError          **err)
{
    uint16_t        msglen;

    /* Short-circuit on no message available */
    if (!fbuf->msgbase) return TRUE;

    /* Close current set */
    fBufAppendSetClose(fbuf);

    /* Store message length */
    msglen = g_htons(fbuf->cp - fbuf->msgbase);
    memcpy(fbuf->msgbase + 2, &msglen, sizeof(msglen));

#if FB_DEBUG_WR
    fBufDebugHex("emit", fbuf->buf, fbuf->cp - fbuf->msgbase);
#endif
#if FB_DEBUG_LWR
    fprintf(stderr, "emit %u (%04x)\n",
            fbuf->cp - fbuf->msgbase, fbuf->cp - fbuf->msgbase);
#endif

    /* Hand the message content to the exporter */
    if (!fbExportMessage(fbuf->exporter, fbuf->buf,
                         fbuf->cp - fbuf->msgbase, err))
        return FALSE;

    /* Increment next record sequence number */
    fbSessionSetSequence(fbuf->session, fbSessionGetSequence(fbuf->session) + fbuf->rc);

    /* Rewind message */
    fBufRewind(fbuf);

    /* All done */
    return TRUE;
}


/**
 * fBufGetExporter
 *
 *
 *
 *
 *
 */
fbExporter_t    *fBufGetExporter(
    fBuf_t          *fbuf)
{
    return fbuf->exporter;
}


/**
 * fBufSetExporter
 *
 *
 *
 *
 *
 */
void            fBufSetExporter(
    fBuf_t          *fbuf,
    fbExporter_t    *exporter)
{
    if (fbuf->collector) {
        fbCollectorFree(fbuf->collector);
        fbuf->collector = NULL;
    }

    if (fbuf->exporter) {
        fbExporterFree(fbuf->exporter);
    }

    fbuf->exporter = exporter;
    fbSessionSetTemplateBuffer(fbuf->session, fbuf);
    fBufRewind(fbuf);
}


/**
 * fBufAllocForExport
 *
 *
 *
 *
 *
 */
fBuf_t          *fBufAllocForExport(
    fbSession_t     *session,
    fbExporter_t    *exporter)
{
    fBuf_t          *fbuf = NULL;

    /* Allocate a new buffer */
    fbuf = g_slice_new0(fBuf_t);

    /* Store reference to session */
    fbuf->session = session;

    /* Set up exporter */
    fBufSetExporter(fbuf, exporter);

    /* Buffers are automatic by default */
    fbuf->automatic = TRUE;

    return fbuf;
}


/**
 * fBufSetExportTime
 *
 *
 *
 *
 *
 */
void            fBufSetExportTime(
    fBuf_t          *fbuf,
    uint32_t        extime)
{
    fbuf->extime = extime;
}

/*==================================================================
 *
 * Buffer Consume (Reader) Functions
 *
 *==================================================================*/

#define FB_CHECK_AVAIL(_op_, _size_)                            \
    if (_size_ > FB_REM_MSG(fbuf)) {                            \
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_EOM,         \
                    "End of message %s "                        \
                    "(need %u bytes, %u available)",            \
                    (_op_), (_size_), FB_REM_MSG(fbuf));        \
        return FALSE;                                           \
    }


#if HAVE_ALIGNED_ACCESS_REQUIRED

#define FB_NEXT_U16(_val_) {                                    \
        uint16_t _x;                                            \
        memcpy (&_x, fbuf->cp, sizeof(uint16_t));               \
        (_val_) = g_ntohs(_x);                                  \
        fbuf->cp += sizeof(uint16_t);                           \
    }

#define FB_NEXT_U32(_val_) {                                    \
        uint32_t _x;                                            \
        memcpy (&_x, fbuf->cp, sizeof(uint32_t));               \
        (_val_) = g_ntohl(_x);                                  \
        fbuf->cp += sizeof(uint32_t);                           \
    }
#else

#define FB_NEXT_U16(_val_)                                \
    (_val_)   = g_ntohs(*((uint16_t *)fbuf->cp));         \
    fbuf->cp += sizeof(uint16_t);

#define FB_NEXT_U32(_val_)                                \
    (_val_)   = g_ntohl(*((uint32_t *)fbuf->cp));         \
    fbuf->cp += sizeof(uint32_t);

#endif




/**
 * fBufNextMessage
 *
 *
 *
 *
 *
 */
gboolean fBufNextMessage(
    fBuf_t          *fbuf,
    GError          **err)
{
    size_t          msglen;
    uint16_t        mh_version, mh_len;
    uint32_t        ex_sequence, mh_sequence, mh_domain;

    /* Need a collector */
    g_assert(fbuf->collector);

    /* Clear external template */
    fbuf->ext_tid = 0;
    fbuf->ext_tmpl = NULL;

    /* Rewind the buffer before reading a new message */
    fBufRewind(fbuf);

    /* Read next message from the collector */
    msglen = sizeof(fbuf->buf);
    if (!fbCollectMessage(fbuf->collector, fbuf->buf, &msglen, err))
        return FALSE;

    /* Set the message end pointer */
    fbuf->mep = fbuf->cp + msglen;

#if FB_DEBUG_RD
    fBufDebugHex("read", fbuf->buf, msglen);
#endif
#if FB_DEBUG_LWR
    fprintf(stderr, "read %lu (%04lx)\n", msglen, msglen);
#endif

    /* Make sure we have at least a message header */
    FB_CHECK_AVAIL("reading message header", 16);

#if FB_DEBUG_RD
    fBufDebugBuffer("rmsg", fbuf, 16, FALSE);
#endif

    /* Read and verify version */
    FB_NEXT_U16(mh_version);
    if (mh_version != 0x000A) {
        g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IPFIX,
                    "Illegal IPFIX Message version 0x%04x; "
                    "input is probably not an IPFIX Message stream.",
                    mh_version);
        return FALSE;
    }

    /* Read and verify message length */
    FB_NEXT_U16(mh_len);
    if (mh_len != msglen) {
        if (NULL != fbuf->collector) {
            if (!fbCollectorHasTranslator(fbuf->collector)){
                g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IPFIX,
                            "IPFIX Message length mismatch "
                            "(buffer has %lu, read %u)",
                            msglen, mh_len);
                return FALSE;
                
            }
        } else {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IPFIX,
                        "IPFIX Message length mismatch "
                        "(buffer has %lu, read %u)",
                        msglen, mh_len);
            return FALSE;
        }
    }

    /* Read and store export time */
    FB_NEXT_U32(fbuf->extime);

    /* Read sequence number */
    FB_NEXT_U32(mh_sequence);

    /* Read observation domain ID and reset domain if necessary */
    FB_NEXT_U32(mh_domain);
    fbSessionSetDomain(fbuf->session, mh_domain);

    /* Verify and update sequence number */
    ex_sequence = fbSessionGetSequence(fbuf->session);
    if (ex_sequence != mh_sequence) {
        if (ex_sequence) {
            g_warning("IPFIX Message out of sequence "
                      "(in domain %08x, expected %08x, got %08x)",
                      fbSessionGetDomain(fbuf->session), ex_sequence, mh_sequence);
        }
        fbSessionSetSequence(fbuf->session, mh_sequence);
    }

    /*
     * We successfully read a message header.
     * Set message base pointer to start of message.
     */
    fbuf->msgbase = fbuf->cp - 16;

    return TRUE;
}


/**
 * fBufSkipCurrentSet
 *
 *
 *
 *
 *
 */
static void     fBufSkipCurrentSet(
    fBuf_t          *fbuf)
{
    if (fbuf->setbase) {
        fbuf->cp += FB_REM_SET(fbuf);
        fbuf->setbase = NULL;
        fbuf->sep = NULL;
    }
}


/**
 * fBufNextSetHeader
 *
 *
 *
 *
 *
 */
static gboolean fBufNextSetHeader(
    fBuf_t          *fbuf,
    GError          **err)
{
    uint16_t        set_id, setlen;

    /* May loop over sets if we're missing templates */
    while (1) {
        /* Make sure we have at least a set header */
        FB_CHECK_AVAIL("reading set header", 4);

#if FB_DEBUG_RD
    fBufDebugBuffer("rset", fbuf, 4, FALSE);
#endif

        /* Read set ID */
        FB_NEXT_U16(set_id);

        /* Read set length */
        FB_NEXT_U16(setlen);

        /* Verify set length is legal */
        if (setlen < 4) {
            g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IPFIX,
                        "Illegal IPFIX Set length %hu",
                        setlen);
            return FALSE;
        }

        /* Verify set body fits in the message */
        FB_CHECK_AVAIL("checking set length", setlen - 4);

        /* Set up special set ID or external templates  */
        if (set_id < FB_TID_MIN_DATA) {
            if ((set_id != FB_TID_TS) &&
                (set_id != FB_TID_OTS)) {
                g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IPFIX,
                            "Illegal IPFIX Set ID %04hx", set_id);
                return FALSE;
            }
            fbuf->spec_tid = set_id;
        } else if (!fbuf->ext_tmpl || fbuf->ext_tid != set_id) {
            fbuf->spec_tid = 0;
            fbuf->ext_tid = set_id;
            fbuf->ext_tmpl = fbSessionGetTemplate(fbuf->session, FALSE, set_id, err);
            if (!fbuf->ext_tmpl) {
                if (g_error_matches(*err, FB_ERROR_DOMAIN, FB_ERROR_TMPL)) {
                    /* Merely warn and skip on missing templates */
                    g_warning("Skipping set: %s", (*err)->message);
                    g_clear_error(err);
                    fbuf->setbase = fbuf->cp - 4;
                    fbuf->sep = fbuf->setbase + setlen;
                    fBufSkipCurrentSet(fbuf);
                    continue;
                }
            }
        }

        /*
        * We successfully read a set header.
        * Set set base and end pointers.
        */
        fbuf->setbase = fbuf->cp - 4;
        fbuf->sep = fbuf->setbase + setlen;

        return TRUE;
    }
}


/**
 * fBufConsumeTemplateSet
 *
 *
 *
 *
 *
 */
static gboolean fBufConsumeTemplateSet(
    fBuf_t          *fbuf,
    GError          **err)
{
    uint16_t        mtl, tid, ie_count, scope_count;
    fbTemplate_t    *tmpl;
    fbInfoElement_t ex_ie = FB_IE_NULL;
    int             i;

    /* Calculate minimum template record length based on type */
    /* FIXME handle revocation sets */
    mtl = (fbuf->spec_tid == FB_TID_OTS) ? 6 : 4;

    /* Keep reading until the set contains only padding. */
    while (FB_REM_SET(fbuf) >= mtl) {
        /* Read template ID */
        FB_NEXT_U16(tid);

        /* Read template IE count */
        FB_NEXT_U16(ie_count);

        /* Read scope count if present */
        if (fbuf->spec_tid == FB_TID_OTS) {
            FB_NEXT_U16(scope_count);

            /* Check for illegal scope count */
            if (scope_count == 0) {
                g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IPFIX,
                            "Illegal IPFIX Options Template Scope Count 0");
                return FALSE;
            } else if (scope_count > ie_count) {
                g_set_error(err, FB_ERROR_DOMAIN, FB_ERROR_IPFIX,
                            "Illegal IPFIX Options Template Scope Count "
                            "(scope count %hu, element count %hu)",
                            scope_count, ie_count);
                return FALSE;
            }
        } else {
            scope_count = 0;
        }

        /* Allocate a new template */
        tmpl = fbTemplateAlloc(fbSessionGetInfoModel(fbuf->session));

        /* Add information elements to the template */
        for (i = 0; i < ie_count; i++) {
            /* Read information element specifier from buffer */
            FB_NEXT_U16(ex_ie.num);
            FB_NEXT_U16(ex_ie.len);
            if (ex_ie.num & IPFIX_ENTERPRISE_BIT) {
                ex_ie.num &= ~IPFIX_ENTERPRISE_BIT;
                FB_NEXT_U32(ex_ie.ent);
            } else {
                ex_ie.ent = 0;
            }

            /* Add information element to template */
            if (!fbTemplateAppend(tmpl, &ex_ie, err)) return FALSE;
        }

        /* Set scope count in template */
        if (scope_count) fbTemplateSetOptionsScope(tmpl, scope_count);

        /* Add template to session */
        if (!fbSessionAddTemplate(fbuf->session, FALSE, tid, tmpl, err)) return FALSE;

        /* Templates are records too */
        /* Again, no. */
        /* ++(fbuf->rc); */

#if FB_DEBUG_RD
    fBufDebugBuffer("rtpl", fbuf, tmpl->tmpl_len, TRUE);
#endif
    }

    /* Skip any padding at the end of the set */
    fBufSkipCurrentSet(fbuf);

    /* All done */
    return TRUE;
}


/**
 * fBufNextDataSet
 *
 *
 *
 *
 *
 */
static gboolean fBufNextDataSet(
    fBuf_t          *fbuf,
    GError          **err)
{
    /* May have to consume multiple template sets */
    while (1) {
        /* Read the next set header */
        if (!fBufNextSetHeader(fbuf, err)) {
            return FALSE;
        }

        /* Check to see if we need to consume a template set */
        if (fbuf->spec_tid) {
            if (!fBufConsumeTemplateSet(fbuf, err)) {
                return FALSE;
            }
            continue;
        }

        /* All done. */
        return TRUE;
    }
}


/**
 * fBufGetCollectionTemplate
 *
 *
 *
 *
 *
 */
fbTemplate_t    *fBufGetCollectionTemplate(
    fBuf_t          *fbuf,
    uint16_t        *ext_tid)
{
    if (fbuf->ext_tmpl) {
        if (ext_tid) *ext_tid = fbuf->ext_tid;
    }
    return fbuf->ext_tmpl;
}


/**
 * fBufNextCollectionTemplateSingle
 *
 *
 *
 *
 *
 */
static fbTemplate_t    *fBufNextCollectionTemplateSingle(
    fBuf_t          *fbuf,
    uint16_t        *ext_tid,
    GError          **err)
{
    /* Read a new message if necessary */
    if (!fbuf->msgbase) {
        if (!fBufNextMessage(fbuf, err)) {
            return FALSE;
        }
    }

    /* Skip any padding at end of current data set */
    if (fbuf->setbase &&
        (FB_REM_SET(fbuf) < fbuf->ext_tmpl->ie_len)) {
        fBufSkipCurrentSet(fbuf);
    }

    /* Advance to the next data set if necessary */
    if (!fbuf->setbase) {
        if (!fBufNextDataSet(fbuf, err)) {
            return FALSE;
        }
    }

    return fBufGetCollectionTemplate(fbuf, ext_tid);
}


/**
 * fBufNextCollectionTemplate
 *
 *
 *
 *
 *
 */
fbTemplate_t    *fBufNextCollectionTemplate(
    fBuf_t          *fbuf,
    uint16_t        *ext_tid,
    GError          **err)
{
    fbTemplate_t    *tmpl;

    while (1) {
        /* Attempt single record read */
        if ((tmpl = fBufNextCollectionTemplateSingle(fbuf, ext_tid, err)))
            return tmpl;

        /* Finish the message at EOM */
        if (g_error_matches(*err, FB_ERROR_DOMAIN, FB_ERROR_EOM)) {
            /* Store next expected sequence number */
            fbSessionSetSequence(fbuf->session,
                                 fbSessionGetSequence(fbuf->session) +
                                 fbuf->rc);

            /* Rewind buffer to force next record read
               to consume a new message. */
            fBufRewind(fbuf);

            /* Clear error and try again in automatic mode */
            if (fbuf->automatic) {
                g_clear_error(err);
                continue;
            }
        }

        /* Error. Not EOM or not retryable. Fail. */
        return NULL;
    }
}


/**
 * fBufNextSingle
 *
 *
 *
 *
 *
 */
static gboolean fBufNextSingle(
    fBuf_t          *fbuf,
    uint8_t         *recbase,
    size_t          *recsize,
    GError          **err)
{
    size_t          bufsize;

    /* Buffer must have active internal template */
    g_assert(fbuf->int_tmpl);

    /* Read a new message if necessary */
    if (!fbuf->msgbase) {
        if (!fBufNextMessage(fbuf, err)) {
            return FALSE;
        }
    }

    /* Skip any padding at end of current data set */
    if (fbuf->setbase &&
        (FB_REM_SET(fbuf) < fbuf->ext_tmpl->ie_len)) {
        fBufSkipCurrentSet(fbuf);
    }

    /* Advance to the next data set if necessary */
    if (!fbuf->setbase) {
        if (!fBufNextDataSet(fbuf, err)) {
            return FALSE;
        }
    }

    /* Transcode bytes out of buffer */
    bufsize = FB_REM_SET(fbuf);
    if (!fbTranscode(fbuf, TRUE, fbuf->cp, recbase, &bufsize, recsize, err)) {
        return FALSE;
    }

    /* Advance current record pointer by bytes read */
    fbuf->cp += bufsize;

    /* Increment record count */
    ++(fbuf->rc);

#if FB_DEBUG_RD
    fBufDebugBuffer("rrec", fbuf, bufsize, TRUE);
#endif
    /* Done */
    return TRUE;
}


/**
 * fBufNext
 *
 *
 *
 *
 *
 */
gboolean        fBufNext(
    fBuf_t          *fbuf,
    uint8_t         *recbase,
    size_t          *recsize,
    GError          **err)
{
    while (1) {
        /* Attempt single record read */
        if (fBufNextSingle(fbuf, recbase, recsize, err)) return TRUE;

        /* Finish the message at EOM */
        if (g_error_matches(*err, FB_ERROR_DOMAIN, FB_ERROR_EOM)) {
            /* Store next expected sequence number */
            fbSessionSetSequence(fbuf->session,
                                 fbSessionGetSequence(fbuf->session) +
                                 fbuf->rc);

            /* Rewind buffer to force next record read
               to consume a new message. */
            fBufRewind(fbuf);

            /* Clear error and try again in automatic mode */
            if (fbuf->automatic) {
                g_clear_error(err);
                continue;
            }
        }

        /* Error. Not EOM or not retryable. Fail. */
        return FALSE;

    }
}


/**
 * fBufGetCollector
 *
 *
 *
 *
 *
 */
fbCollector_t   *fBufGetCollector(
    fBuf_t          *fbuf)
{
    return fbuf->collector;
}


/**
 * fBufSetCollector
 *
 *
 *
 *
 *
 */
void            fBufSetCollector(
    fBuf_t          *fbuf,
    fbCollector_t   *collector)
{
    if (fbuf->exporter) {
        fbSessionSetTemplateBuffer(fbuf->session, NULL);
        fbExporterFree(fbuf->exporter);
        fbuf->exporter = NULL;
    }

    if (fbuf->collector) {
        fbCollectorFree(fbuf->collector);
    }

    fbuf->collector = collector;
    fBufRewind(fbuf);
}


/**
 * fBufAllocForCollection
 *
 *
 *
 *
 *
 */
fBuf_t          *fBufAllocForCollection(
    fbSession_t     *session,
    fbCollector_t   *collector)
{
    fBuf_t          *fbuf = NULL;

    /* Allocate a new buffer */
    fbuf = g_slice_new0(fBuf_t);

    /* Store reference to session */
    fbuf->session = session;

    /* Set up collection */
    fBufSetCollector(fbuf, collector);

    /* Buffers are automatic by default */
    fbuf->automatic = TRUE;

    return fbuf;
}


/**
 * fBufGetExportTime
 *
 *
 *
 *
 *
 */
uint32_t        fBufGetExportTime(
    fBuf_t          *fbuf)
{
    return fbuf->extime;
}
